import dataclasses
import functools
import logging
import os
import pathlib
import time
from collections.abc import Callable
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Union

import yaml
from slugify import slugify

from substra.sdk import backends
from substra.sdk import exceptions
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.utils import check_and_format_search_filters
from substra.sdk.utils import check_search_ordering
from substra.sdk.utils import is_valid_uuid

logger = logging.getLogger(__name__)

DEFAULT_RETRY_TIMEOUT = 5 * 60
DEFAULT_BATCH_SIZE = 500
DEFAULT_BACKEND_TYPE = schemas.BackendType.LOCAL_SUBPROCESS

# Temporary output identifiers, to be removed once user can pass its own identifiers
MODEL_OUTPUT_IDENTIFIER = "model"
HEAD_OUTPUT_IDENTIFIER = "local"
TRUNK_OUTPUT_IDENTIFIER = "shared"


def logit(f):
    """Decorator used to log all high-level methods of the Substra client."""

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        logger.debug(f"{f.__name__}: call")
        ts = time.time()
        error = None
        try:
            return f(*args, **kwargs)
        except Exception as e:
            error = e.__class__.__name__
            raise
        finally:
            # add a log even if the function raises an exception
            te = time.time()
            elapsed = (te - ts) * 1000
            logger.info(f"{f.__name__}: done in {elapsed:.2f}ms; error={error}")

    return wrapper


def _upper_slug(s: str) -> str:
    return slugify(s, separator="_").upper()


def _get_env_var(attribute_name: str, client_name: str) -> str:
    """
    Returns the environment variable associated with a given attribute for a given Client,
    identified by its client_name
    """
    return f"SUBSTRA_{_upper_slug(client_name)}_{_upper_slug(attribute_name)}"


@dataclasses.dataclass
class SettingValue:
    value: Any
    origin: Literal["code", "environment_variable", "config_file", "default"]


def _get_config_value(
    *,
    attribute_name: str,
    code_values: Dict[str, Any],
    client_name: Optional[str],
    client_config_dict: Dict[str, str],
    mapping: Dict[str, Callable],
) -> SettingValue:
    """
    For a given attribute, return the value to be used in the Client configuration.
    The order of precedence is: values defined by the user in the code, environment variables, values read from the
    configuration file. If the attribute is not set, the value returned is None (and the origin is set to "default").

    Args:
        attribute_name (str): Name of the attribute for which we want to get the value to use
        code_values (dict): Dictionary containing the values read from the code (or None)
        client_name (str or None): Name of the client; it is used to determine which env var to use and which
            sub-dictionary of the configuration file to consider; if None, only the values defined in the code
            are considered.
        client_config_dict (dict): Content of the configuration file
        mapping (dict): Mapping associating a callable to each attribute; usually used for type conversion, but could be
            used for more advanced processing

    Returns:
        SettingValue: Object containing the value and its origin (code, environment_variable, config_file or default)
    """

    # Some values are boolean that might be set to False, so we check explicitly "is not None"
    # Note that explicitly passing "None" in the code is equivalent to omitting it.
    # (code_values will actually always contain all the keys, with None values for keys that were omitted)
    if code_values.get(attribute_name) is not None:
        return SettingValue(value=code_values[attribute_name], origin="code")

    # if client_name is not given, we cannot look through env var and config files
    if not client_name:
        return SettingValue(value=None, origin="default")

    env_var_name = _get_env_var(attribute_name, client_name)
    # In environment variables, though, we never get "None"
    if env_var_name in os.environ:
        return SettingValue(value=mapping[attribute_name](os.environ[env_var_name]), origin="environment_variable")

    if attribute_name in client_config_dict:
        return SettingValue(value=mapping[attribute_name](client_config_dict[attribute_name]), origin="config_file")

    return SettingValue(value=None, origin="default")


def get_client_configuration(
    *,
    config_file: Optional[pathlib.Path],
    client_name: Optional[str],
    code_values: Dict[str, Any],
) -> Dict[str, SettingValue]:
    """
    Get the unified configuration values for a given client, out of three possible sources.
    The order of precedence is: values defined by the user in the code, environment variables, values read from the
    configuration file. If the attribute is not set, the value returned is None (and the origin is set to "default").

    Args:
        config_file (Path or None): Path to the yaml configuration file, if existing
        client_name (str or None): Name of the client; it is used to determine which env var to use and which
            sub-dictionary of the configuration file to consider; if None, only the values defined in the code
            are considered.
        code_values (dict): Value passed by the user directly in the code when instantiating the `Client`

    Returns:
     A dictionary associating each attribute with its `SettingValue` (containing the actual value and its origin)

    """
    if config_file and config_file.exists():
        config_dict = yaml.safe_load(config_file.read_text(encoding=None))
    else:
        config_dict = {}

    values_mapping = {
        "url": str,
        "token": str,
        "username": str,
        "password": str,
        "retry_timeout": int,
        "insecure": bool,
        "backend_type": schemas.BackendType,
    }
    return {
        attribute_name: _get_config_value(
            attribute_name=attribute_name,
            code_values=code_values,
            client_name=client_name,
            client_config_dict=config_dict.get(client_name, {}),
            mapping=values_mapping,
        )
        for attribute_name in values_mapping
    }


class Client:
    """Create a client.
    Defaults to a subprocess client, suitable for development purpose.
    Configuration can be passed in the code, through environment variables or through a configuration file.
    The order of precedence is: values defined by the user in the code, environment variables, values read from the
    configuration file. If the attribute is not set, the value returned is None (and the origin is set to "default").
    In order to use configuration values not explicitly defined in the code, the parameter `client_name` must
    not be None.

    Args:
        client_name (str, optional): Name of the client.
            Used to load relevant environment variables and select the right dictionary in the configuration file.
            Defaults to None.
        configuration_file (path, optional): Path to te configuration file.
            `client_name` must be defined too.
            Defaults to None.
        url (str, optional): URL of the Substra platform.
            Mandatory to connect to a Substra platform. If no URL is given, all assets are created locally.
            Can be set to "" to remove any previously defined URL (in a configuration file or environment variable).
            Defaults to None.
        token (str, optional): Token to authenticate to the Substra platform.
            If no token is given, and a `username`/ `password` pair is provided,  the Client will try to authenticate
            using 'login' function. It's always possible to generate a new token later by making another call to the
            `login` function.
            Defaults to None.
        username (str, optional): Username to authenticate to the Substra platform.
            Used in conjunction with a password to generate a token if not given, using the `login` function.

            If using username/password, you should use a context manager to ensure the session terminates as intended:
            ```
            with Client(username, password) as client:
               ...
            ```
            Not stored.
            Defaults to None.
        password (str, optional): Password to authenticate to the Substra platform.
            Used in conjunction with a username to generate a token if not given, using the `login` function.
            Not stored.
            Defaults to None.
        retry_timeout (int, optional): Number of seconds before attempting a retry call in case of timeout.
            Defaults to 5 minutes.
        insecure (bool, optional): If True, the client can call a not-certified backend.
            This is for development purposes.
            Defaults to False.
        backend_type (schemas.BackendType, optional): Which backend type to use.
            Possible values are `remote`, `docker` and `subprocess`.
            Defaults to `subprocess`.
            In `remote` mode, assets are registered on a deployed platform which also executes the tasks.
            In `subprocess` or `docker` mode, if no URL is given then all assets are created locally and tasks are
            executed locally. If a URL is given then the mode is a hybrid one: new assets are
            created locally but can access assets from the deployed Substra platform. The platform is in read-only mode
            and tasks are executed locally.
    """

    def __init__(
        self,
        *,
        client_name: Optional[str] = None,
        configuration_file: Optional[pathlib.Path] = None,
        url: Optional[str] = None,
        token: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        retry_timeout: Optional[int] = None,
        insecure: Optional[bool] = None,
        backend_type: Optional[schemas.BackendType] = None,
    ):
        # The value "" (which is Falsy) is used to bypass configuration file
        if configuration_file is None:
            configuration_file = os.getenv("SUBSTRA_CLIENTS_CONFIGURATION_FILE_PATH")

            if configuration_file:
                logger.info(
                    "Configuration file path set from env var SUBSTRA_CLIENTS_CONFIGURATION_FILE_PATH: "
                    f"'{configuration_file}'"
                )

        if configuration_file and not client_name:
            raise exceptions.ConfigurationInfoError(
                "Configuration file cannot be used because no `client_name` was given."
            )

        if configuration_file:
            configuration_file = pathlib.Path(configuration_file)

        code_values = {
            "url": url,
            "token": token,
            "username": username,
            "password": password,
            "retry_timeout": retry_timeout,
            "insecure": insecure,
            "backend_type": backend_type,
        }
        config_dict = get_client_configuration(
            client_name=client_name, config_file=configuration_file, code_values=code_values
        )
        if config_dict["backend_type"].value is None:
            logger.info("No backend type specified, defaulting to subprocess")
            backend_type = DEFAULT_BACKEND_TYPE
        else:
            backend_type = config_dict["backend_type"].value
            logger.info(f"Backend type set to {backend_type}, " f"as defined in {config_dict['backend_type'].origin}")

        self._url = config_dict["url"].value
        if self._url:
            logger.info(
                f"The parameter `url` has been set to {self._url} " f"as defined in {config_dict['url'].origin} "
            )
        else:
            logger.info("No URL has been given, all assets will be created locally.")

        self._token = config_dict["token"].value
        if self._token:
            logger.info(f"Token read from {config_dict['token'].origin}")

        self._retry_timeout = (
            config_dict["retry_timeout"].value
            if config_dict["retry_timeout"].value is not None
            else DEFAULT_RETRY_TIMEOUT
        )
        logger.info(
            f"The parameter `retry_timeout` has been set to {self._retry_timeout} "
            f"as defined in {config_dict['retry_timeout'].origin} "
        )

        self._insecure = config_dict["insecure"].value if config_dict["insecure"].value is not None else False

        logger.info(
            f"The parameter `insecure` has been set to {self._insecure} "
            f"as defined in {config_dict['insecure'].origin} "
        )

        self._backend = self._get_backend(backend_type)
        if (
            self._url
            and self._token is None
            and not (config_dict["username"].value is None or config_dict["password"].value is None)
        ):
            logger.info(
                f"No token provided, getting one using username set in {config_dict['username'].origin} "
                f"and password set in {config_dict['password'].origin}"
            )
            self._token = self.login(config_dict["username"].value, config_dict["password"].value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.logout()

    def __del__(self):
        self.logout()

    def _get_backend(self, backend_type: schemas.BackendType):
        # Three possibilities:
        # - remote: get a remote backend
        # - subprocess/docker and no url is defined: fully local backend
        # - subprocess/docker and url is defined: local backend that connects to
        #                           a deployed backend (read-only)
        if backend_type == schemas.BackendType.REMOTE:
            return backends.get(
                schemas.BackendType.REMOTE,
                url=self._url,
                insecure=self._insecure,
                token=self._token,
                retry_timeout=self._retry_timeout,
            )
        if backend_type in [schemas.BackendType.LOCAL_DOCKER, schemas.BackendType.LOCAL_SUBPROCESS]:
            backend = None
            if self._url:
                logger.warning(
                    f"Backend mode {backend_type} given with a URL is a hybrid mode. The client will connect to the "
                    "platform in read-only."
                )
                backend = backends.get(
                    schemas.BackendType.REMOTE,
                    url=self._url,
                    insecure=self._insecure,
                    token=self._token,
                    retry_timeout=self._retry_timeout,
                )
            return backends.get(
                backend_type,
                backend,
            )
        raise exceptions.SDKException(
            f"Unknown value for the execution mode: {backend_type}, "
            f"valid values are: {schemas.BackendType.__members__.values()}"
        )

    @property
    def temp_directory(self):
        """Temporary directory for storing assets in local mode.
        Deleted when the client is deleted.
        """
        if isinstance(self._backend, backends.Local):
            return self._backend.temp_directory

    @property
    def backend_mode(self) -> schemas.BackendType:
        """Get the backend mode.

        Returns:
            str: Backend mode
        """
        return self._backend.backend_mode

    @logit
    def login(self, username, password):
        """Login to a remote server."""
        if not self._backend:
            raise exceptions.SDKException("No backend found")
        self._token = self._backend.login(username, password)
        return self._token

    @logit
    def logout(self) -> None:
        """
        Log out from a remote server, if Client.login was used
        (otherwise, nothing happens)
        """
        if not self._backend:
            raise exceptions.SDKException("No backend found")
        self._backend.logout()
        self._token = None

    @staticmethod
    def _get_spec(asset_type, data):
        if isinstance(data, asset_type):
            return data
        return asset_type(**data)

    @logit
    def add_data_sample(self, data: Union[dict, schemas.DataSampleSpec], local: bool = True) -> str:
        """Create a new data sample asset and return its key.

        Args:
            data (Union[dict, schemas.DataSampleSpec]): data sample to add. If it is a dict,
                it must follow the [DataSampleSpec schema](sdk_schemas.md#DataSampleSpec).
            local (bool, optional): If `local` is true, `path` must refer to a directory located
                on the local filesystem. The file content will be transferred to the server
                through an HTTP query, so this mode should be used for relatively small files
                (<10mo).

                If `local` is false, `path` must refer to a directory located on the server
                filesystem. This directory must be accessible (readable) by the server.  This
                mode is well suited for all kind of file sizes. Defaults to True.

        Returns:
            str: key of the data sample
        """
        spec = self._get_spec(schemas.DataSampleSpec, data)
        if spec.paths:
            raise ValueError("data: invalid 'paths' field")
        if not spec.path:
            raise ValueError("data: missing 'path' field")
        spec_options = {
            "local": local,
        }
        return self._backend.add(spec, spec_options=spec_options)

    @logit
    def add_data_samples(
        self,
        data: Union[dict, schemas.DataSampleSpec],
        local: bool = True,
    ) -> List[str]:
        """Create many data sample assets and return  a list of keys.

        Create multiple data samples through a single HTTP request.
        This method is well suited for adding multiple small files only. For adding a
        large amount of data it is recommended to add them one by one. It allows a
        better control in case of failures.

        Args:
            data (Union[dict, schemas.DataSampleSpec]): data samples to add. If it is a dict,
                it must follow the [DataSampleSpec schema](sdk_schemas.md#DataSampleSpec).
                The `paths` in the data dictionary must be a list of paths where each path
                points to a directory representing one data sample.
            local (bool, optional):  Please refer to the method `Client.add_data_sample`.
                Defaults to True.

        Returns:
            List[str]: List of the data sample keys
        """
        spec = self._get_spec(schemas.DataSampleSpec, data)
        if spec.path:
            raise ValueError("data: invalid 'path' field")
        if not spec.paths:
            raise ValueError("data: missing 'paths' field")
        spec_options = {
            "local": local,
        }
        return self._backend.add(
            spec,
            spec_options=spec_options,
        )

    @logit
    def add_dataset(self, data: Union[dict, schemas.DatasetSpec]):
        """Create new dataset asset and return its key.

        Args:
            data (Union[dict, schemas.DatasetSpec]): If it is a dict, it must have the same
                keys as specified in [schemas.DatasetSpec](sdk_schemas.md#DatasetSpec).

        Returns:
            str: Key of the dataset
        """
        spec = self._get_spec(schemas.DatasetSpec, data)
        return self._backend.add(spec)

    @logit
    def add_function(self, data: Union[dict, schemas.FunctionSpec]) -> str:
        """Create new function asset.

        Args:
            data (Union[dict, schemas.FunctionSpec]): If it is a dict, it must have the same keys
                as specified in [schemas.FunctionSpec](sdk_schemas.md#FunctionSpec).

        Returns:
            str: Key of the function
        """
        spec = self._get_spec(schemas.FunctionSpec, data)
        return self._backend.add(spec)

    @logit
    def add_task(self, data: Union[dict, schemas.TaskSpec]) -> str:
        """Create new task asset.

        Args:
            data (Union[dict, schemas.TaskSpec]): If it is a dict, it must have the same
                keys as specified in [schemas.TaskSpec](sdk_schemas.md#TaskSpec).

        Returns:
            str: Key of the asset
        """
        spec = self._get_spec(schemas.TaskSpec, data)
        return self._backend.add(spec)

    @logit
    def add_compute_plan(
        self,
        data: Union[dict, schemas.ComputePlanSpec],
        auto_batching: bool = True,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> models.ComputePlan:
        """Create new compute plan asset.

        Args:
            data (Union[dict, schemas.ComputePlanSpec]): If it is a dict, it must have the same
                keys as specified in [schemas.ComputePlanSpec](sdk_schemas.md#ComputePlanSpec).
            auto_batching (bool, optional): Set 'auto_batching' to False to upload all the tasks of
                the compute plan at once. Defaults to True.
            batch_size (int, optional): If 'auto_batching' is True, change `batch_size` to define
                the number of tasks uploaded in each batch (default 500).

        Returns:
            models.ComputePlan: Created compute plan
        """
        spec = self._get_spec(schemas.ComputePlanSpec, data)
        spec_options = {
            "auto_batching": auto_batching,
            "batch_size": batch_size,
        }

        if not is_valid_uuid(spec.key):
            raise exceptions.ComputePlanKeyFormatError(
                "The compute plan key has to respect the UUID format. You can use the uuid library to generate it. \
            Example: compute_plan_key=str(uuid.uuid4())"
            )

        return self._backend.add(spec, spec_options=spec_options)

    @logit
    def get_function(self, key: str) -> models.Function:
        """Get function by key, the returned object is described
        in the [models.Function](sdk_models.md#Function) model"""
        return self._backend.get(schemas.Type.Function, key)

    @logit
    def get_compute_plan(self, key: str) -> models.ComputePlan:
        """Get compute plan by key, the returned object is described
        in the [models.ComputePlan](sdk_models.md#ComputePlan) model"""
        return self._backend.get(schemas.Type.ComputePlan, key)

    @logit
    def get_performances(self, key: str, *, wait_completion: bool = False) -> models.Performances:
        """Get the compute plan performances by key, the returned object is described
        in the [models.Performances](sdk_models.md#Performances) and easily convertible
        to pandas dataframe. You can wait for compute task to finish by setting
        `wait_completion = True`

        Example:
            ```python
            perf = client.get_performances(cp_key)
            df = pd.DataFrame(perf.model_dump())
            print(df)
            ```
        """
        if wait_completion:
            self.wait_compute_plan(key)
        performances = self._backend.get_performances(key)
        return performances

    @logit
    def get_dataset(self, key: str) -> models.Dataset:
        """Get dataset by key, the returned object is described
        in the [models.Dataset](sdk_models.md#Dataset) model"""
        return self._backend.get(schemas.Type.Dataset, key)

    @logit
    def get_task(self, key: str) -> models.Task:
        """Get task by key, the returned object is described
        in the [models.Task](sdk_models.md#Task) model"""
        return self._backend.get(schemas.Type.Task, key)

    @logit
    def get_logs(self, task_key: str) -> str:
        """Get task logs by task key, the returned object is a string
        containing the logs.

        Logs are only available for tasks that experienced an execution failure.
        Attempting to retrieve logs for tasks in any other states or for non-existing
        tasks will result in a NotFound error.
        """
        return self._backend.download_logs(task_key, destination_file=None)

    @logit
    def get_model(self, key: str) -> models.OutModel:
        return self._backend.get(schemas.Type.Model, key)

    @logit
    def list_model(self, filters: dict = None, ascending: bool = False) -> List[models.OutModel]:
        """List models.

        The ``filters`` argument is a dictionnary, with those possible keys:\n
            key (list[str]): list model with given keys.\n
            compute_task_key (list[str]): list model produced by this compute task.\n
            owner (list[str]): list model with given owners.\n
            permissions (list[str]): list models which can be used by any of the listed nodes. Remote mode only.\n

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            ascending (bool, optional): Sorts results by oldest creation_date first. Default False (descending order).

        Returns:
            models.OutModel the returned object is described in the [models.OutModel](sdk_models.md#OutModel) model"""
        return self._list(schemas.Type.Model, filters, "creation_date", ascending)

    @logit
    def get_data_sample(self, key: str) -> models.DataSample:
        """Get data sample by key, the returned object is described
        in the [models.Datasample](sdk_models.md#DataSample) model"""
        return self._backend.get(schemas.Type.DataSample, key)

    def _list(
        self, asset_type, filters: dict = None, order_by: str = None, ascending: bool = False, paginated: bool = True
    ):
        filters = check_and_format_search_filters(asset_type, filters)
        check_search_ordering(order_by)
        return self._backend.list(asset_type, filters, order_by, ascending)

    @logit
    def list_function(self, filters: dict = None, ascending: bool = False) -> List[models.Function]:
        """List functions.

        The ``filters`` argument is a dictionary, with those possible keys:\n
            key (List[str]): list function with given keys.\n
            name (str): list function with name partially matching given string. Remote mode only.\n
            owner (List[str]): list function with given owners.\n
            metadata (dict)
                {
                    "key": str # the key of the metadata to filter on
                    "type": "is", "contains" or "exists" # the type of query that will be used
                    "value": str # the value that the key must be (if type is "is") or contain (if type if "contains")
                }: list function matching provided conditions in metadata.\n
            permissions (List[str]): list function which can be used by any of the listed nodes. Remote mode only.
            compute_plan_key (str): list function that are in the given compute plan. Remote mode only.
            dataset_key (str): list function linked or using this dataset. Remote mode only.
            data_sample_key (List[str]): list function linked or that used this data sample(s). Remote mode only.

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            ascending (bool, optional): Sorts results by oldest creation_date first. Default False (descending order).

        Returns:
            models.Function: the returned object is described in the [models.Function](sdk_models.md#Function) model

        """

        return self._list(schemas.Type.Function, filters, "creation_date", ascending)

    @logit
    def list_compute_plan(
        self, filters: dict = None, order_by: str = "creation_date", ascending: bool = False
    ) -> List[models.ComputePlan]:
        """List compute plans.

         The ``filters`` argument is a dictionary, with those possible keys:\n
            key (List[str]): list compute plans with listed keys.\n
            name (str): list compute plans with name partially matching given string. Remote mode only.\n
            owner (List[str]): list compute plans with listed owners.\n
            worker (List[str]): list compute plans which ran on listed workers. Remote mode only.\n
            status (List[str]): list compute plans with given status.
                The possible values are the values of `substra.models.ComputePlanStatus`
            metadata (dict)
                {
                    "key": str # the key of the metadata to filter on
                    "type": "is", "contains" or "exists" # the type of query that will be used
                    "value": str # the value that the key must be (if type is "is") or contain (if type if "contains")
                }: list compute plans matching provided conditions in metadata. Remote mode only.
            function_key (str): list compute plans that used the given function. Remote mode only.
            dataset_key (str): list compute plans linked or using this dataset. Remote mode only.
            data_sample_key (List[str]): list compute plans linked or that used this data sample(s). Remote mode only.

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            order_by (str, optional): Field to sort results by.
                Possible values: `creation_date`, `start_date`, `end_date`. Default creation_date.
            ascending (bool, optional): Sorts results on order_by by ascending order. Default False (descending order).

        Returns:
            models.ComputePlan: the returned object is described
        in the [models.ComputePlan](sdk_models.md#ComputePlan) model
        """

        return self._list(schemas.Type.ComputePlan, filters, order_by, ascending)

    @logit
    def list_data_sample(self, filters: dict = None, ascending: bool = False) -> List[models.DataSample]:
        """List data samples.

            The ``filters`` argument is a dictionary, with those possible keys:\n
                key (List[str]): list data samples with listed keys.
                owner (List[str]): list data samples with listed owners.
                compute_plan_key (str): list data samples that are in the given compute plan. Remote mode only.
                function_key (str): list data samples that used the given function. Remote mode only.
                dataset_key (str): list data samples linked or using this dataset. Remote mode only.

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            ascending (bool, optional): Sorts results on order_by by ascending order. Default False (descending order).

        Returns:
            models.DataSample: the returned object is described in the
                [models.DataSample](sdk_models.md#DataSample) model
        """
        return self._list(schemas.Type.DataSample, filters, "creation_date", ascending)

    @logit
    def list_dataset(self, filters: dict = None, ascending: bool = False) -> List[models.Dataset]:
        """List datasets.

         The ``filters`` argument is a dictionary, with those possible keys:\n
            key (List[str]): list dataset with given keys.\n
            name (str): list dataset with name partially matching given string. Remote mode only.\n
            owner (List[str]): list dataset with given owners.\n
            metadata (dict)
                {
                    "key": str # the key of the metadata to filter on
                    "type": "is", "contains" or "exists" # the type of query that will be used
                    "value": str # the value that the key must be (if type is "is") or contain (if type if "contains")
                }: list dataset matching provided conditions in metadata.\n
            permissions (List[str]) : list dataset which can be used by any of the listed nodes. Remote mode only.
            compute_plan_key (str): list dataset that are in the given compute plan. Remote mode only.
            function_key (str): list dataset that used the given function. Remote mode only.
            data_sample_key (List[str]): list dataset linked or that used this data sample(s). Remote mode only.

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            ascending (bool, optional): Sorts results by oldest creation_date first. Default False (descending order).

        Returns:
            models.Dataset: the returned object is described
        in the [models.Dataset](sdk_models.md#Dataset) model


        """

        return self._list(schemas.Type.Dataset, filters, "creation_date", ascending)

    @logit
    def list_task(
        self, filters: dict = None, order_by: str = "creation_date", ascending: bool = False
    ) -> List[models.Task]:
        """List tasks.

        The ``filters`` argument is a dictionary, with those possible keys:\n
            key (List[str]): list tasks with listed keys.\n
            owner (List[str]): list tasks with listed owners.\n
            worker (List[str]): list tasks which ran on listed workers. Remote mode only.\n
            rank (List[int]): list tasks which are at given ranks.\n
            status (List[str]): list tasks with given status.
                The possible values are the values of `substra.models.ComputeTaskStatus`
            metadata (dict)
                {
                    "key": str # the key of the metadata to filter on
                    "type": "is", "contains" or "exists" # the type of query that will be used
                    "value": str # the value that the key must be (if type is "is") or contain (if type if "contains")
                }: list tasks matching provided conditions in metadata. Remote mode only.
            compute_plan_key (str): list tasks that are in the given compute plan. Remote mode only.
            function_key (str): list tasks that used the given function. Remote mode only.

        Args:
            filters (dict, optional): List of key values pair to filter on. Default None.
            order_by (str, optional): Field to sort results by.
                Possible values: `creation_date`, `start_date`, `end_date`. Default creation_date.
            ascending (bool, optional): Sorts results on order_by by ascending order. Default False (descending order).

        Returns:
            models.Task: the returned object is described in the
                [models.Task](sdk_models.md#Task) model


        """
        return self._list(schemas.Type.Task, filters, order_by, ascending)

    @logit
    def list_task_input_assets(self, key: str) -> List[models.InputAsset]:
        """List input assets for a specific task, the returned object is described
        in the [models.InputAsset](sdk_models.md#InputAsset) model"""
        return self._backend.list_task_input_assets(key)

    @logit
    def list_task_output_assets(self, key: str, *, wait_completion: bool = False) -> List[models.OutputAsset]:
        """List output assets for a specific task, the returned object is described
        in the [models.OutputAsset](sdk_models.md#OutputAsset) model. You can wait
        for compute task to finish by setting `wait_completion = True`"""
        if wait_completion:
            self.wait_task(key)
        return self._backend.list_task_output_assets(key)

    @logit
    def get_task_output_asset(self, key: str, identifier: str, *, wait_completion: bool = False) -> models.OutputAsset:
        """Get an output asset for a specific task with a defined identifier, the returned object is described
        in the [models.OutputAsset](sdk_models.md#OutputAsset) model. You can wait
        for compute task to finish by setting `wait_completion = True`"""
        if wait_completion:
            self.wait_task(key)
        return self._backend.get_task_output_asset(key, identifier)

    @logit
    def list_organization(self, *args, **kwargs) -> List[models.Organization]:
        """List organizations, the returned object is described
        in the [models.Organization](sdk_models.md#Organization) model"""
        # results are not paginated for organizations
        return self._backend.list(schemas.Type.Organization, paginated=False)

    @logit
    def update_function(self, key: str, name: str):
        spec = self._get_spec(schemas.UpdateFunctionSpec, {"name": name})
        self._backend.update(key, spec)
        return

    @logit
    def add_compute_plan_tasks(
        self,
        key: str,
        tasks: Union[dict, schemas.UpdateComputePlanTasksSpec],
        auto_batching: bool = True,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> models.ComputePlan:
        """Update compute plan.

        Args:
            key (str): Compute plan key
            tasks (Union[dict, schemas.UpdateComputePlanTasksSpec]): If it is a dict,
                it must have the same keys as specified in
                [schemas.UpdateComputePlanTasksSpec](sdk_schemas.md#UpdateComputePlanTasksSpec).
            auto_batching (bool, optional): Set 'auto_batching' to False to upload all
                the tasks of the compute plan at once. Defaults to True.
            batch_size (int, optional): If 'auto_batching' is True, change `batch_size`
                to define the number of tasks uploaded in each batch (default 500).

        Returns:
            models.ComputePlan: updated compute plan, as described in the
            [models.ComputePlan](sdk_models.md#ComputePlan) model
        """
        if isinstance(tasks, dict):
            tasks["key"] = key
        spec = self._get_spec(schemas.UpdateComputePlanTasksSpec, tasks)
        spec_options = {
            "auto_batching": auto_batching,
            "batch_size": batch_size,
        }
        return self._backend.add_compute_plan_tasks(spec, spec_options=spec_options)

    @logit
    def update_compute_plan(self, key: str, name: str):
        spec = self._get_spec(schemas.UpdateComputePlanSpec, {"name": name})
        self._backend.update(key, spec)
        return

    @logit
    def update_dataset(self, key: str, name: str):
        spec = self._get_spec(schemas.UpdateDatasetSpec, {"name": name})
        self._backend.update(key, spec)
        return

    @logit
    def link_dataset_with_data_samples(
        self,
        dataset_key: str,
        data_sample_keys: List[str],
    ) -> List[str]:
        """Link dataset with data samples."""
        return self._backend.link_dataset_with_data_samples(dataset_key, data_sample_keys)

    @logit
    def download_dataset(self, key: str, destination_folder: str) -> pathlib.Path:
        """Download data manager resource.
        Download opener script in destination folder.

        Args:
            key (str): Dataset key to download
            destination_folder (str): Destination folder

        Returns:
            pathlib.Path: Path of the downloaded dataset
        """
        return pathlib.Path(
            self._backend.download(
                schemas.Type.Dataset,
                "opener.storage_address",
                key,
                os.path.join(destination_folder, "opener.py"),
            )
        )

    @logit
    def download_function(self, key: str, destination_folder: str) -> pathlib.Path:
        """Download function resource.
        Download function package in destination folder.

        Args:
            key (str): Function key to download
            destination_folder (str): Destination folder

        Returns:
            pathlib.Path: Path of the downloaded function
        """

        return pathlib.Path(
            self._backend.download(
                schemas.Type.Function,
                "archive.storage_address",
                key,
                os.path.join(destination_folder, "function.tar.gz"),
            )
        )

    @logit
    def download_model(self, key: str, destination_folder) -> pathlib.Path:
        """Download model to destination file.

        This model was saved using the 'save_model' function of the class.
        To load and use the model, please refer to the 'load_model' and 'predict' functions of the
        class.

        Args:
            key (str): Model key to download
            destination_folder (str): Destination folder

        Returns:
            pathlib.Path: Path of the downloaded model
        """
        return pathlib.Path(self._backend.download_model(key, os.path.join(destination_folder, f"model_{key}")))

    @logit
    def download_model_from_task(self, task_key: str, identifier: str, folder: os.PathLike) -> pathlib.Path:
        """Download task model to destination file.

        This model was saved using the 'save_model' function of the class.
        To load and use the model, please refer to the 'load_model' and 'predict' functions of the
        class.

        Args:
            task_key (str): Task key to download
            identifier(str): output identifier
            folder (os.PathLike): Destination folder

        Returns:
            pathlib.Path: Path of the downloaded model
        """
        task_output = self._backend.get_task_output_asset(task_key, identifier)
        model = task_output.asset
        return self.download_model(model.key, folder)

    @logit
    def download_logs(self, task_key: str, folder: str) -> str:
        """Download the execution logs of a failed task to a destination file.

        The logs are saved in the folder to a file named 'task_logs_{task_key}.txt'.

        Logs are only available for tasks that experienced an execution failure.
        Attempting to retrieve logs for tasks in any other states or for non-existing
        tasks will result in a NotFound error.

        Args:
            task_key: the key of the task that produced the logs
            folder: the destination directory

        Returns:
            str: The logs as a str
        """
        return self._backend.download_logs(task_key, os.path.join(folder, f"task_logs_{task_key}.txt"))

    @logit
    def describe_function(self, key: str) -> str:
        """Get function description."""
        return self._backend.describe(schemas.Type.Function, key)

    @logit
    def describe_dataset(self, key: str) -> str:
        """Get dataset description."""
        return self._backend.describe(schemas.Type.Dataset, key)

    @logit
    def organization_info(self) -> models.OrganizationInfo:
        """Get organization information."""
        return self._backend.organization_info()

    @logit
    def cancel_compute_plan(self, key: str) -> None:
        """Cancel execution of compute plan. Nothing is returned by this method"""
        self._backend.cancel_compute_plan(key)

    @logit
    def wait_compute_plan(
        self, key: str, *, timeout: Optional[float] = None, polling_period: float = 2.0, raise_on_failure: bool = True
    ) -> models.ComputePlan:
        """Wait for the execution of the given compute plan to finish.

        It is considered finished when the status is done, failed or cancelled.

        Args:
            key (str): the key of the compute plan to wait for
            timeout (float, optional): maximum time to wait, in seconds. If set to None, will hang until completion.
            polling_period (float): time to wait between two checks, in seconds. Defaults to 2.0.
            raise_on_failure (bool): whether to raise an exception if the execution fails. Defaults to True.

        Returns:
            models.ComputePlan: the compute plan after completion

        Raises:
            exceptions.FutureFailureError: The compute plan failed or have been cancelled.
            exceptions.FutureTimeoutError: The compute plan took more than the duration set in the timeout to complete.
                Not raised when `timeout == None`
        """
        asset_getter = self.get_compute_plan
        status_failed = models.ComputePlanStatus.failed.value
        status_canceled = models.ComputePlanStatus.canceled.value
        statuses_stopped = (
            models.ComputePlanStatus.done.value,
            models.ComputePlanStatus.failed.value,
            models.ComputePlanStatus.canceled.value,
        )

        return self._wait(
            key=key,
            asset_getter=asset_getter,
            polling_period=polling_period,
            raise_on_failure=raise_on_failure,
            status_canceled=status_canceled,
            status_failed=status_failed,
            statuses_stopped=statuses_stopped,
            timeout=timeout,
        )

    @logit
    def wait_task(
        self, key: str, *, timeout: Optional[float] = None, polling_period: float = 2.0, raise_on_failure: bool = True
    ) -> models.Task:
        """Wait for the execution of the given task to finish.

        It is considered finished when the status is done, failed or cancelled.

        Args:
            key (str): the key of the task to wait for.
            timeout (float, optional): maximum time to wait, in seconds. If set to None, will hang until completion.
            polling_period (float): time to wait between two checks, in seconds. Defaults to 2.0.
            raise_on_failure (bool): whether to raise an exception if the execution fails. Defaults to True.

        Returns:
            models.Task: the task after completion

         Raises:
            exceptions.FutureFailureError: The task failed or have been cancelled.
            exceptions.FutureTimeoutError: The task took more than the duration set in the timeout to complete.
                Not raised when `timeout == None`
        """
        asset_getter = self.get_task
        status_canceled = models.ComputeTaskStatus.canceled.value
        status_failed = models.ComputeTaskStatus.failed.value
        statuses_stopped = (models.ComputeTaskStatus.done.value, models.ComputeTaskStatus.canceled.value)
        return self._wait(
            key=key,
            asset_getter=asset_getter,
            polling_period=polling_period,
            raise_on_failure=raise_on_failure,
            status_canceled=status_canceled,
            status_failed=status_failed,
            statuses_stopped=statuses_stopped,
            timeout=timeout,
        )

    @logit
    def wait_function(
        self, key: str, *, timeout: Optional[float] = None, polling_period: float = 1.0, raise_on_failure: bool = True
    ) -> models.Function:
        """Wait for the build of the given function to finish.

        It is considered finished when the status is ready, failed or cancelled.

        Args:
            key (str): the key of the task to wait for.
            timeout (float, optional): maximum time to wait, in seconds. If set to None, will hang until completion.
            polling_period (float): time to wait between two checks, in seconds. Defaults to 2.0.
            raise_on_failure (bool): whether to raise an exception if the execution fails. Defaults to True.

        Returns:
            models.Function: the function once built

         Raises:
            exceptions.FutureFailureError: The function build failed or have been cancelled.
            exceptions.FutureTimeoutError: The function took more than the duration set in the timeout to build.
                Not raised when `timeout == None`
        """
        asset_getter = self.get_function
        status_canceled = models.FunctionStatus.canceled.value
        status_failed = models.FunctionStatus.failed.value
        statuses_stopped = (
            models.FunctionStatus.ready.value,
            models.FunctionStatus.canceled.value,
            models.FunctionStatus.failed.value,
        )
        return self._wait(
            key=key,
            asset_getter=asset_getter,
            polling_period=polling_period,
            raise_on_failure=raise_on_failure,
            status_canceled=status_canceled,
            status_failed=status_failed,
            statuses_stopped=statuses_stopped,
            timeout=timeout,
        )

    def _wait(
        self,
        *,
        key: str,
        asset_getter,
        polling_period: float,
        raise_on_failure: bool,
        status_failed: str,
        status_canceled: str,
        statuses_stopped: Sequence[str],
        timeout: Optional[float] = None,
    ):
        tstart = time.time()
        while True:
            asset = asset_getter(key)

            if asset.status in statuses_stopped:
                break

            if asset.status == models.ComputeTaskStatus.failed.value and asset.error_type is not None:
                # when dealing with a failed task, wait for the error_type field of the task to be set
                # i.e. wait for the registration of the failure report
                break

            if timeout and time.time() - tstart > timeout:
                raise exceptions.FutureTimeoutError(f"Future timeout on {asset}")

            time.sleep(polling_period)

        if raise_on_failure and asset.status == status_failed:
            raise exceptions.FutureFailureError(f"Future execution failed on {asset}")

        if raise_on_failure and asset.status == status_canceled:
            raise exceptions.FutureFailureError(f"Future execution canceled on {asset}")

        return asset



import argparse
import inspect
import sys
from pathlib import Path

import docstring_parser

from substra import Client
from substra.sdk.utils import retry_on_exception

MODULE_LIST = [Client, retry_on_exception]

KEYWORDS = ["Args", "Returns", "Yields", "Raises", "Example"]


def generate_function_help(fh, asset):
    """Write the description of a function"""
    fh.write(f"# {asset.__name__}\n")
    signature = str(inspect.signature(asset))
    fh.write("```text\n")
    fh.write(f"{asset.__name__}{signature}")
    fh.write("\n```")
    fh.write("\n\n")
    # Write the docstring
    docstring = inspect.getdoc(asset)
    docstring = docstring_parser.parse(inspect.getdoc(asset))
    fh.write(f"{docstring.short_description}\n")
    if docstring.long_description:
        fh.write(f"{docstring.long_description}\n")
    # Write the arguments as a list
    if len(docstring.params) > 0:
        fh.write("\n**Arguments:**\n")
    for param in docstring.params:
        type_and_optional = ""
        if param.type_name or param.is_optional is not None:
            text_optional = "required"
            if param.is_optional:
                text_optional = "optional"
            type_and_optional = f"({param.type_name}, {text_optional})"
        fh.write(f" - `{param.arg_name} {type_and_optional}`: {param.description}\n")
    # Write everything else as is
    for param in [
        meta_param for meta_param in docstring.meta if not isinstance(meta_param, docstring_parser.DocstringParam)
    ]:
        fh.write(f"\n**{param.args[0].title()}:**\n")
        if len(param.args) > 1:
            for extra_param in param.args[1:]:
                fh.write(f"\n - `{extra_param}`: ")
        fh.write(f"{param.description}\n")


def generate_properties_help(fh, public_methods):
    properties = [(f_name, f_method) for f_name, f_method in public_methods if isinstance(f_method, property)]
    for f_name, f_method in properties:
        fh.write(f"## {f_name}\n")
        fh.write("_This is a property._  \n")
        fh.write(f"{f_method.__doc__}\n")


def generate_help(fh):
    for asset in MODULE_LIST:
        if inspect.isclass(asset):  # Class
            public_methods = [
                (f_name, f_method) for f_name, f_method in inspect.getmembers(asset) if not f_name.startswith("_")
            ]
            generate_function_help(fh, asset)
            generate_properties_help(fh, public_methods)
            for _, f_method in public_methods:
                if not isinstance(f_method, property):
                    fh.write("#")  # Title for the methods are one level below
                    generate_function_help(fh, f_method)
        elif callable(asset):
            generate_function_help(fh, asset)


def write_help(path):
    with path.open("w") as fh:
        generate_help(fh)


if __name__ == "__main__":
    default_path = Path(__file__).resolve().parents[1] / "references" / "sdk.md"

    parser = argparse.ArgumentParser()
    parser.add_argument("--output-path", type=str, default=str(default_path.resolve()), required=False)

    args = parser.parse_args(sys.argv[1:])
    write_help(Path(args.output_path))



import logging
import pathlib
import shutil
import tempfile
import typing

from substra.sdk import exceptions
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.backends.local import db
from substra.sdk.backends.remote import backend

logger = logging.getLogger(__name__)


class DataAccess:
    """Data access layer.

    This is an intermediate layer between the backend and the local/remote data access.
    """

    def __init__(self, remote_backend: typing.Optional[backend.Remote], local_worker_dir: pathlib.Path):
        self._db = db.get_db()
        self._remote = remote_backend
        self._tmp_dir = tempfile.TemporaryDirectory(prefix=str(local_worker_dir) + "/")

    @property
    def tmp_dir(self):
        return pathlib.Path(self._tmp_dir.name)

    def is_local(self, key: str, type_: schemas.Type):
        try:
            self._db.get(type_, key)
            return True
        except exceptions.NotFound:
            return False

    def _get_asset_content_filename(self, type_):
        if type_ == schemas.Type.Function:
            filename = "function.tar.gz"
            field_name = "archive"

        elif type_ == schemas.Type.Dataset:
            filename = "opener.py"
            field_name = "opener"

        else:
            raise ValueError(f"Cannot download this type of asset {type_}")

        return filename, field_name

    def login(self, username, password):
        if self._remote:
            self._remote.login(username, password)

    def logout(self):
        if self._remote:
            self._remote.logout()

    def add(self, asset):
        return self._db.add(asset)

    def remote_download(self, asset_type, url_field_path, key, destination):
        self._remote.download(asset_type, url_field_path, key, destination)

    def remote_download_model(self, key, destination_file):
        self._remote.download_model(key, destination_file)

    def get_remote_description(self, asset_type, key):
        return self._remote.describe(asset_type, key)

    def get_with_files(self, type_: schemas.Type, key: str):
        """Get the asset with files on the local disk for execution.
        This does not load the description as it is not required for execution.
        """
        try:
            # Try to find the asset locally
            return self._db.get(type_, key)
        except exceptions.NotFound:
            if self._remote is not None:
                # if not found, try remotely
                filename, field_name = self._get_asset_content_filename(type_)
                asset = self._remote.get(type_, key)
                tmp_directory = self.tmp_dir / key
                asset_path = tmp_directory / filename

                if not tmp_directory.exists():
                    pathlib.Path.mkdir(tmp_directory)

                    self._remote.download(
                        type_,
                        field_name + ".storage_address",
                        key,
                        asset_path,
                    )

                attr = getattr(asset, field_name)
                attr.storage_address = asset_path
                return asset
            raise

    def get(self, type_, key: str):
        try:
            # Try to find the asset locally
            return self._db.get(type_, key)
        except exceptions.NotFound:
            if self._remote is not None:
                return self._remote.get(type_, key)
            raise

    def get_performances(self, key: str) -> models.Performances:
        """Get the performances of a given compute. Return models.Performances() object
        easily convertible to dict, filled by the performances data of done tasks that output a performance.
        """
        compute_plan = self.get(schemas.Type.ComputePlan, key)
        list_tasks = self.list(
            schemas.Type.Task,
            filters={"compute_plan_key": [key]},
            order_by="rank",
            ascending=True,
        )

        performances = models.Performances()

        for task in list_tasks:
            if task.status == models.ComputeTaskStatus.done:
                function = self.get(schemas.Type.Function, task.function.key)
                perf_identifiers = [
                    output.identifier for output in function.outputs if output.kind == schemas.AssetKind.performance
                ]
                outputs = self.list(
                    schemas.Type.OutputAsset, {"compute_task_key": task.key, "identifier": perf_identifiers}
                )
                for output in outputs:
                    performances.compute_plan_key.append(compute_plan.key)
                    performances.compute_plan_tag.append(compute_plan.tag)
                    performances.compute_plan_status.append(compute_plan.status)
                    performances.compute_plan_start_date.append(compute_plan.start_date)
                    performances.compute_plan_end_date.append(compute_plan.end_date)
                    performances.compute_plan_metadata.append(compute_plan.metadata)

                    performances.worker.append(task.worker)
                    performances.task_key.append(task.key)
                    performances.task_rank.append(task.rank)
                    try:
                        round_idx = int(task.metadata.get("round_idx"))
                    except TypeError:
                        round_idx = None
                    performances.round_idx.append(round_idx)
                    performances.identifier.append(output.identifier)
                    performances.performance.append(output.asset)

        return performances

    def list(
        self, type_: str, filters: typing.Dict[str, typing.List[str]], order_by: str = None, ascending: bool = False
    ):
        """Joins the results of the [local db](substra.sdk.backends.local.db.list) and the
        [remote db](substra.sdk.backends.rest_client.list) in hybrid mode.
        """

        local_assets = self._db.list(type_=type_, filters=filters, order_by=order_by, ascending=ascending)

        remote_assets = []
        if self._remote:
            try:
                remote_assets = self._remote.list(
                    asset_type=type_, filters=filters, order_by=order_by, ascending=ascending
                )
            except Exception as e:
                logger.info(
                    f"Could not list assets from the remote platform:\n{e}. \
                    \nIf you are not logged to a remote platform, ignore this message."
                )
        return local_assets + remote_assets

    def save_file(self, file_path: typing.Union[str, pathlib.Path], key: str):
        """Copy file or directory into the local temp dir to mimick
        the remote backend that saves the files given by the user.
        """
        tmp_directory = self.tmp_dir / key
        tmp_file = tmp_directory / pathlib.Path(file_path).name

        if not tmp_directory.exists():
            pathlib.Path.mkdir(tmp_directory)

        if tmp_file.exists():
            raise exceptions.AlreadyExists(f"File {tmp_file.name} already exists for asset {key}", 409)
        elif pathlib.Path(file_path).is_file():
            shutil.copyfile(file_path, tmp_file)
        elif pathlib.Path(file_path).is_dir():
            shutil.copytree(file_path, tmp_file)
        else:
            raise exceptions.InvalidRequest(f"Could not copy {file_path}", 400)
        return tmp_file

    def update(self, asset):
        self._db.update(asset)
        return



import os
import stat
import zipfile


class ZipFile(zipfile.ZipFile):
    """Override Zipfile to ensure unix file permissions are preserved.

    This is due to a python bug:
    https://bugs.python.org/issue15795

    Workaround from:
    https://stackoverflow.com/questions/39296101/python-zipfile-removes-execute-permissions-from-binaries
    """

    def extract(self, member, path=None, pwd=None):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        if path is None:
            path = os.getcwd()

        ret_val = self._extract_member(member, path, pwd)
        attr = member.external_attr >> 16
        os.chmod(ret_val, attr)
        return ret_val

    def extractall(self, path=None, members=None, pwd=None):
        self._sanity_check()
        super().extractall(path, members, pwd)

    def _sanity_check(self):
        """Check that the archive does not attempt to traverse path.

        This is inspired by TarSafe: https://github.com/beatsbears/tarsafe
        """
        for zipinfo in self.infolist():
            if self._is_traversal_attempt(zipinfo):
                raise Exception(f"Attempted directory traversal for member: {zipinfo.filename}")
            if self._is_symlink(zipinfo):
                raise Exception(f"Unsupported symlink for member: {zipinfo.filename}")

    def _is_traversal_attempt(self, zipinfo: zipfile.ZipInfo) -> bool:
        base_directory = os.getcwd()
        zipfile_path = os.path.abspath(os.path.join(base_directory, zipinfo.filename))
        if not zipfile_path.startswith(base_directory):
            return True
        return False

    def _is_symlink(self, zipinfo: zipfile.ZipInfo) -> bool:
        return zipinfo.external_attr & stat.S_IFLNK << 16 == stat.S_IFLNK << 16



import copy
import logging
import shutil
import typing
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import NoReturn
from typing import Optional

from tqdm.auto import tqdm

import substra
from substra.sdk import compute_plan as compute_plan_module
from substra.sdk import exceptions
from substra.sdk import fs
from substra.sdk import graph
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.backends import base
from substra.sdk.backends.local import compute
from substra.sdk.backends.local import dal

logger = logging.getLogger(__name__)

_MAX_LEN_KEY_METADATA = 50
_MAX_LEN_VALUE_METADATA = 100


class Local(base.BaseBackend):
    org_counter = 1

    def __init__(self, backend, backend_type, *args, **kwargs):
        self._local_worker_dir = Path.cwd() / "local-worker"
        self._local_worker_dir.mkdir(exist_ok=True)

        self._execution_mode = backend_type
        # create a store to abstract the db
        self._db = dal.DataAccess(backend, local_worker_dir=self._local_worker_dir)
        self._worker = compute.Worker(
            self._db,
            local_worker_dir=self._local_worker_dir,
            debug_spawner=self._execution_mode,
        )

        self._org_id = f"MyOrg{Local.org_counter}MSP"
        self._org_name = f"MyOrg{Local.org_counter}MSP"
        Local.org_counter += 1
        self._db.add(
            models.Organization(
                id=self._org_id,
                is_current=False,  # all orgs share the same db, catch and change this value at the "get"
                creation_date=self.__now(),
            )
        )

    def __now(self):
        """Return the current date in the iso format"""
        return datetime.now()

    @property
    def temp_directory(self):
        """Get the temporary directory where the assets are saved.
        The directory is deleted at the end of the execution."""
        return self._db.tmp_dir

    @property
    def backend_mode(self) -> schemas.BackendType:
        """Get the backend mode"""
        return self._execution_mode

    def login(self, username, password):
        self._db.login(username, password)

    def logout(self):
        self._db.logout()

    def get(self, asset_type, key):
        return self._db.get(asset_type, key)

    def get_task_output_asset(self, compute_task_key: str, identifier: str) -> models.OutputAsset:
        outputs = self._db.list(
            schemas.Type.OutputAsset, {"identifier": identifier, "compute_task_key": compute_task_key}
        )
        if len(outputs) == 0:
            raise exceptions.TaskAssetNotFoundError(compute_task_key=compute_task_key, identifier=identifier)
        elif len(outputs) > 1:
            raise exceptions.TaskAssetMultipleFoundError(compute_task_key=compute_task_key, identifier=identifier)

        return outputs[0]

    def list_task_output_assets(self, compute_task_key: str) -> List[models.OutputAsset]:
        outputs = self._db.list(schemas.Type.OutputAsset, {"compute_task_key": compute_task_key})

        return outputs

    def list_task_input_assets(self, compute_task_key: str) -> List[models.InputAsset]:
        inputs = self._db.list(schemas.Type.InputAsset, {"compute_task_key": compute_task_key})

        return inputs

    def get_performances(self, key):
        performances = self._db.get_performances(key)

        return performances

    def list(
        self,
        asset_type: schemas.Type,
        filters: Dict[str, List[str]] = None,
        order_by: str = None,
        ascending: bool = False,
        paginated: bool = True,
    ) -> List[models._Model]:
        """List the assets

        Args:
            paginated (bool, optional): True if server response is expected to be paginated.
                Ignored for local backend (responses are never paginated)
            others: cf [dal](substra.sdk.backends.local.dal.list)

        Returns:
            List[models._Model]: List of results
        """
        results = self._db.list(type_=asset_type, filters=filters, order_by=order_by, ascending=ascending)
        if asset_type == schemas.Type.Organization:
            for i, item in enumerate(results):
                if item.id == self._org_id:
                    my_org = copy.copy(item)  # do not change the value in the db
                    my_org.is_current = True
                    results[i] = my_org
        return results

    @staticmethod
    def _check_metadata(metadata: Optional[Dict[str, str]]):
        if metadata is not None:
            if any([len(key) > _MAX_LEN_KEY_METADATA for key in metadata]):
                raise exceptions.InvalidRequest("The key in metadata cannot be more than 50 characters", 400)
            if any([len(str(value)) > _MAX_LEN_VALUE_METADATA or len(str(value)) == 0 for value in metadata.values()]):
                raise exceptions.InvalidRequest("Values in metadata cannot be empty or more than 100 characters", 400)
            if any("__" in key for key in metadata):
                raise exceptions.InvalidRequest(
                    '"__" cannot be used in a metadata key, please use simple underscore instead', 400
                )

    def _check_data_samples(self, spec: schemas.TaskSpec):
        """Get all the data samples from the db once and run sanity checks on them"""

        dataset_input_ref = [
            input_ref.asset_key
            for input_ref in spec.inputs
            if input_ref.identifier == schemas.StaticInputIdentifier.opener
        ]
        data_sample_keys = [
            input_ref.asset_key
            for input_ref in spec.inputs
            if input_ref.identifier == schemas.StaticInputIdentifier.datasamples
        ]

        if len(dataset_input_ref) == 1 and len(data_sample_keys) > 0:
            data_manager_key = dataset_input_ref[0]
            data_samples = self.list(schemas.Type.DataSample, filters={"key": data_sample_keys})

            if self._db.is_local(data_manager_key, schemas.Type.Dataset):
                data_manager = self._db.get(schemas.Type.Dataset, data_manager_key)
                if data_manager.owner != spec.worker:
                    raise substra.exceptions.InvalidRequest(
                        f"The task worker must be the organization that contains the data: {data_manager.owner}", 400
                    )

            # Check that we've got all the data_samples
            if len(data_samples) != len(data_sample_keys):
                raise substra.exceptions.InvalidRequest(
                    "Could not get all the data_samples in the database with the given data_sample_keys", 400
                )

            self._check_same_data_manager(data_manager_key, data_samples)

    def _check_same_data_manager(self, data_manager_key, data_samples):
        """Check that all data samples are linked to this data manager"""
        # If the dataset is remote: the backend does not return the datasets
        # linked to each sample, so no check (already done in the backend).
        if self._db.is_local(data_manager_key, schemas.Type.Dataset):
            same_data_manager = all(
                [data_manager_key in data_sample.data_manager_keys for data_sample in (data_samples or [])]
            )
            if not same_data_manager:
                raise substra.exceptions.InvalidRequest("A data_sample does not belong to the same dataManager", 400)

    def __compute_permissions(self, permissions):
        """Compute the permissions

        If the permissions are private, the active organization is
        in the authorized ids.
        """
        updated_permissions = copy.deepcopy(permissions)
        owner = self._org_id
        if updated_permissions.public:
            updated_permissions.authorized_ids = []
        elif not updated_permissions.public and owner not in updated_permissions.authorized_ids:
            updated_permissions.authorized_ids.append(owner)
        return updated_permissions

    def __add_compute_plan(
        self,
        task_count,
    ):
        key = schemas._Spec.compute_key()
        compute_plan = models.ComputePlan(
            key=key,
            creation_date=self.__now(),
            start_date=self.__now(),
            status=models.ComputePlanStatus.created,
            tag="",
            name=key,
            task_count=task_count,
            waiting_for_executor_slot=task_count,
            metadata=dict(),
            owner="local",
            delete_intermediary_models=False,
        )

        return self._db.add(compute_plan)

    def __create_compute_plan_from_task(self, spec, in_tasks):
        # compute plan and rank
        if spec.compute_plan_key is not None:
            compute_plan = self._db.get(schemas.Type.ComputePlan, spec.compute_plan_key)
            compute_plan_key = compute_plan.key
            if spec.rank is not None:
                # Use the rank given by the user
                rank = spec.rank
            else:
                if len(in_tasks) == 0:
                    rank = 0
                else:
                    rank = 1 + max(
                        [in_task.rank for in_task in in_tasks if in_task.compute_plan_key == compute_plan_key]
                    )

            # Add to the compute plan
            compute_plan.task_count += 1
            compute_plan.waiting_executor_slot_count += 1
            compute_plan.status = models.ComputePlanStatus.created

        elif not spec.compute_plan_key and (spec.rank == 0 or spec.rank is None):
            # Create a compute plan
            compute_plan = self.__add_compute_plan(task_count=1)
            rank = 0
            compute_plan_key = compute_plan.key
        else:
            raise substra.sdk.exceptions.InvalidRequest("invalid inputs, a new ComputePlan should have a rank 0", 400)

        return compute_plan_key, rank

    def __execute_compute_plan(self, spec, compute_plan, visited, tasks, spec_options):
        with tqdm(
            total=len(visited),
            desc="Compute plan progress",
        ) as progress_bar:
            for id_, _ in sorted(visited.items(), key=lambda item: item[1]):
                task_spec = tasks.get(id_)
                if not task_spec:
                    continue

                self.add(
                    key=task_spec.key,
                    spec=task_spec,
                    spec_options=spec_options,
                )

                progress_bar.update()

        return compute_plan

    def _add_function(self, key, spec, spec_options=None):
        self._check_metadata(spec.metadata)
        permissions = self.__compute_permissions(spec.permissions)
        function_file_path = self._db.save_file(spec.file, key)
        function_description_path = self._db.save_file(spec.description, key)

        function = models.Function(
            key=key,
            creation_date=self.__now(),
            name=spec.name,
            owner=self._org_id,
            permissions={
                "process": {
                    "public": permissions.public,
                    "authorized_ids": permissions.authorized_ids,
                },
            },
            archive={"checksum": fs.hash_file(function_file_path), "storage_address": function_file_path},
            description={
                "checksum": fs.hash_file(function_description_path),
                "storage_address": function_description_path,
            },
            metadata=spec.metadata if spec.metadata else dict(),
            inputs=_schemas_list_to_models_list(spec.inputs, models.FunctionInput),
            outputs=_schemas_list_to_models_list(spec.outputs, models.FunctionOutput),
            status=models.FunctionStatus.ready,
        )
        return self._db.add(function)

    def _add_dataset(self, key, spec, spec_options=None):
        self._check_metadata(spec.metadata)

        permissions = self.__compute_permissions(spec.permissions)
        logs_permission = self.__compute_permissions(spec.logs_permission)

        dataset_file_path = self._db.save_file(spec.data_opener, key)
        dataset_description_path = self._db.save_file(spec.description, key)
        asset = models.Dataset(
            key=key,
            creation_date=self.__now(),
            owner=self._org_id,
            name=spec.name,
            permissions={
                "process": {
                    "public": permissions.public,
                    "authorized_ids": permissions.authorized_ids,
                },
            },
            data_sample_keys=[],
            opener={"checksum": fs.hash_file(dataset_file_path), "storage_address": dataset_file_path},
            description={
                "checksum": fs.hash_file(dataset_description_path),
                "storage_address": dataset_description_path,
            },
            metadata=spec.metadata if spec.metadata else dict(),
            logs_permission=logs_permission.model_dump(),
        )
        return self._db.add(asset)

    def _add_data_sample(self, key, spec, spec_options=None):
        if len(spec.data_manager_keys) == 0:
            raise exceptions.InvalidRequest("Please add at least one data manager for the data sample", 400)
        datasets = [self._db.get(schemas.Type.Dataset, dataset_key) for dataset_key in spec.data_manager_keys]

        data_sample = models.DataSample(
            key=key,
            creation_date=self.__now(),
            owner=self._org_id,
            path=spec.path,
            data_manager_keys=spec.data_manager_keys,
        )
        data_sample = self._db.add(data_sample)

        # update dataset(s) accordingly
        for dataset in datasets:
            samples_list = dataset.data_sample_keys
            if data_sample.key not in samples_list:
                samples_list.append(data_sample.key)

        return data_sample

    def _add_data_samples(self, spec, spec_options=None):
        data_samples = []
        for path in spec.paths:
            data_sample_spec = schemas.DataSampleSpec(
                path=path,
                data_manager_keys=spec.data_manager_keys,
            )
            key = data_sample_spec.compute_key()
            data_sample = self._add_data_sample(
                key,
                data_sample_spec,
                spec_options,
            )
            data_samples.append(data_sample)
        return data_samples

    def _add_compute_plan(self, spec: schemas.ComputePlanSpec, spec_options: dict = None):
        self._check_metadata(spec.metadata)
        # Get all the tasks and their dependencies
        task_graph, tasks = compute_plan_module.get_dependency_graph(spec)

        # Define the rank of each task
        visited = graph.compute_ranks(node_graph=task_graph)

        compute_plan = models.ComputePlan(
            key=spec.key,
            creation_date=self.__now(),
            start_date=self.__now(),
            tag=spec.tag or "",
            name=spec.name,
            status=models.ComputePlanStatus.created,
            metadata=spec.metadata or dict(),
            task_count=0,
            waiting_builder_slot_count=0,
            building_count=0,
            waiting_parent_tasks_count=0,
            waiting_executor_slot_count=0,
            executing_count=0,
            canceled_count=0,
            failed_count=0,
            done_count=0,
            owner=self._org_id,
        )
        compute_plan = self._db.add(compute_plan)

        # go through the tasks sorted by rank
        compute_plan = self.__execute_compute_plan(spec, compute_plan, visited, tasks, spec_options)
        return compute_plan

    def _add_task(self, key, spec, spec_options=None):
        self._check_metadata(spec.metadata)
        self._check_data_samples(spec)
        function = self._db.get(schemas.Type.Function, spec.function_key)

        _warn_on_transient_outputs(spec.outputs)

        in_task_keys = list({inputref.parent_task_key for inputref in (spec.inputs or []) if inputref.parent_task_key})
        in_tasks = [self._db.get(schemas.Type.Task, in_task_key) for in_task_key in in_task_keys]
        compute_plan_key, rank = self.__create_compute_plan_from_task(spec=spec, in_tasks=in_tasks)

        task = models.Task(
            key=key,
            creation_date=self.__now(),
            function=function,
            owner=self._org_id,
            worker=spec.worker,
            compute_plan_key=compute_plan_key,
            rank=rank,
            inputs=_schemas_list_to_models_list(spec.inputs, models.InputRef),
            outputs=_output_from_spec(spec.outputs),
            tag=spec.tag or "",
            # TODO: the waiting status should be more granular now
            status=models.ComputeTaskStatus.waiting_for_executor_slot,
            metadata=spec.metadata if spec.metadata else dict(),
        )

        task = self._db.add(task)
        self._worker.schedule_task(task)
        return task

    def _check_inputs_outputs(self, spec, function_key):
        function = self._db.get(schemas.Type.Function, function_key)
        spec_inputs = spec.inputs or []
        spec_outputs = spec.outputs or dict()

        error_msg = ""
        for function_input in function.inputs:
            spec_values = [x for x in spec_inputs if x.identifier == function_input.identifier]
            if len(spec_values) == 0 and not function_input.optional:
                error_msg += f"  - The input {function_input.identifier} is not optional according to the function.\n"
            if len(spec_values) > 1 and not function_input.multiple:
                error_msg += f"  - The input {function_input.identifier} is not multiple according to the function.\n"

        if not set([x.identifier for x in spec_inputs]).issubset(set([x.identifier for x in function.inputs])):
            error_msg += "  - There are keys in the spec inputs that have not been declared in the function inputs.\n"

        spec_output_keys = set([x for x in spec_outputs])
        function_output_keys = set([x.identifier for x in function.outputs])
        if not spec_output_keys == function_output_keys:
            error_msg += (
                "  - The identifiers in the spec outputs and the function outputs must match:"
                f"{spec_output_keys.symmetric_difference(function_output_keys)}.\n"
            )

        for output_performance in (x for x in function.outputs if x.kind == schemas.AssetKind.performance):
            if not spec_outputs[output_performance.identifier].permissions.public:
                error_msg += '  - invalid task output "performance": a PERFORMANCE output should be public.'
                break

        if len(error_msg) > 0:
            raise exceptions.InvalidRequest(f"There are errors in the inputs / outputs fields:\n{error_msg}", "OE0101")

    def add(self, spec, spec_options=None, key=None):
        function_key = getattr(spec, "function_key", None)
        if function_key is not None:
            self._check_inputs_outputs(spec=spec, function_key=function_key)

        # find dynamically the method to call to create the asset
        method_name = f"_add_{spec.__class__.type_.value}"
        if spec.is_many():
            method_name += "s"
        add_asset = getattr(self, method_name)
        if spec.is_many():
            assets = add_asset(spec, spec_options)
            return [asset.key for asset in assets]
        else:
            if spec.__class__.type_ == schemas.Type.ComputePlan:
                compute_plan = add_asset(spec, spec_options)
                return compute_plan
            else:
                key = key or spec.compute_key()
                add_asset(key, spec, spec_options)
                return key

    def link_dataset_with_data_samples(self, dataset_key, data_sample_keys) -> List[str]:
        dataset = self._db.get(schemas.Type.Dataset, dataset_key)
        data_samples = []
        for key in data_sample_keys:
            data_sample = self._db.get(schemas.Type.DataSample, key)
            if dataset_key not in data_sample.data_manager_keys:
                data_sample.data_manager_keys.append(dataset_key)
                dataset.data_sample_keys.append(key)
            else:
                logger.warning(f"Data sample already in dataset: {key}")
            data_samples.append(data_sample)
        return data_sample_keys

    def download(self, asset_type, url_field_path, key, destination_file):
        if self._db.is_local(key, asset_type):
            asset = self._db.get(type_=asset_type, key=key)
            # Get the field containing the path to the file.
            file_path = asset
            for field in url_field_path.split("."):
                file_path = getattr(file_path, field, None)
            # Copy the file to the destination file
            shutil.copyfile(file_path, destination_file)
        else:
            self._db.remote_download(asset_type, url_field_path, key, destination_file)

        return destination_file

    def download_model(self, key, destination_file):
        if self._db.is_local(key, schemas.Type.Model):
            asset = self._db.get(type_=schemas.Type.Model, key=key)
            shutil.copyfile(asset.address.storage_address, destination_file)
        else:
            self._db.remote_download_model(key, destination_file)

        return destination_file

    def download_logs(self, task_key: str, destination_file: str = None) -> NoReturn:
        raise NotImplementedError("Logs of tasks ran in local mode are not accessible")

    def describe(self, asset_type, key):
        if self._db.is_local(key, schemas.Type.Dataset):
            asset = self._db.get(type_=asset_type, key=key)
            if not hasattr(asset, "description") or not asset.description:
                raise ValueError("This element does not have a description.")
            with open(asset.description.storage_address, "r", encoding="utf-8") as f:
                return f.read()
        else:
            return self._db.get_remote_description(asset_type, key)

    def organization_info(self):
        return models.OrganizationInfo(
            host="http://fakeurl.fake",
            organization_id=self._org_id,
            organization_name=self._org_name,
            config=models.OrganizationInfoConfig(
                model_export_enabled=True,
            ),
            channel="channel1",
            version=substra.__version__,
            orchestrator_version="fake",
        )

    def cancel_compute_plan(self, key):
        # Execution is synchronous in the local backend so this
        # function does not make sense.
        raise NotImplementedError

    def update(self, key, spec, spec_options=None):
        asset_type = spec.__class__.type_
        asset = self.get(asset_type, key)
        data = asset.model_dump()
        data.update(spec.model_dump())
        updated_asset = models.SCHEMA_TO_MODEL[asset_type](**data)
        self._db.update(updated_asset)
        return

    def add_compute_plan_tasks(self, spec: schemas.UpdateComputePlanTasksSpec, spec_options: dict = None):
        key = spec.key
        compute_plan = self._db.get(schemas.Type.ComputePlan, key)
        # Get all the new tasks and their dependencies
        task_graph, tasks = compute_plan_module.get_dependency_graph(spec)

        # Get the rank of all the tasks already in the compute plan
        filters = {"compute_plan_key": [key]}
        cp_tasks = self.list(
            asset_type=schemas.Type.Task,
            filters=filters,
        )
        visited = dict()
        for task in cp_tasks:
            visited[task.key] = task.rank

        # Update the task graph with the tasks already in the CP
        task_graph.update({k: [] for k in visited})
        visited = graph.compute_ranks(node_graph=task_graph, ranks=visited)

        compute_plan = self.__execute_compute_plan(spec, compute_plan, visited, tasks, spec_options)
        return compute_plan


def _output_from_spec(outputs: Dict[str, schemas.ComputeTaskOutputSpec]) -> Dict[str, models.ComputeTaskOutput]:
    """Convert a list of schemas.ComputeTaskOuput to a list of models.ComputeTaskOutput"""
    return {
        identifier: models.ComputeTaskOutput(
            permissions=models.Permissions(process=output.permissions.model_dump()),
            transient=output.is_transient,
            value=None,
        )
        # default isNone (= outputs are not computed yet)
        for identifier, output in outputs.items()
    }


def _schemas_list_to_models_list(inputs: Any, model: Any) -> Any:
    if not inputs:
        return []
    return [model.model_validate(input_schema.model_dump()) for input_schema in inputs]


def _warn_on_transient_outputs(outputs: typing.Dict[str, schemas.ComputeTaskOutputSpec]):
    for _, output in outputs.items():
        if output.is_transient:
            warnings.warn("`transient=True` is ignored in local mode", stacklevel=1)



import abc
import enum
import re
from datetime import datetime
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pydantic
from pydantic import AnyUrl
from pydantic import ConfigDict
from pydantic import DirectoryPath
from pydantic import FilePath
from pydantic.fields import Field

from substra.sdk import schemas

# The remote can return an URL or an empty string for paths
UriPath = Union[FilePath, AnyUrl, str]
CAMEL_TO_SNAKE_PATTERN = re.compile(r"(.)([A-Z][a-z]+)")
CAMEL_TO_SNAKE_PATTERN_2 = re.compile(r"([a-z0-9])([A-Z])")


class MetadataFilterType(str, enum.Enum):
    is_equal = "is"
    contains = "contains"
    exists = "exists"


class ComputeTaskStatus(str, enum.Enum):
    """Status of the task"""

    unknown = "STATUS_UNKNOWN"
    building = "STATUS_BUILDING"
    executing = "STATUS_EXECUTING"
    done = "STATUS_DONE"
    failed = "STATUS_FAILED"
    waiting_for_executor_slot = "STATUS_WAITING_FOR_EXECUTOR_SLOT"
    waiting_for_parent_tasks = "STATUS_WAITING_FOR_PARENT_TASKS"
    waiting_for_builder_slot = "STATUS_WAITING_FOR_BUILDER_SLOT"
    canceled = "STATUS_CANCELED"


class ComputePlanStatus(str, enum.Enum):
    """Status of the compute plan"""

    unknown = "PLAN_STATUS_UNKNOWN"
    doing = "PLAN_STATUS_DOING"
    done = "PLAN_STATUS_DONE"
    failed = "PLAN_STATUS_FAILED"
    created = "PLAN_STATUS_CREATED"
    canceled = "PLAN_STATUS_CANCELED"


class FunctionStatus(str, enum.Enum):
    """Status of the function"""

    unknown = "FUNCTION_STATUS_UNKNOWN"
    waiting = "FUNCTION_STATUS_WAITING"
    building = "FUNCTION_STATUS_BUILDING"
    ready = "FUNCTION_STATUS_READY"
    failed = "FUNCTION_STATUS_FAILED"
    canceled = "FUNCTION_STATUS_CANCELED"


class TaskErrorType(str, enum.Enum):
    """Types of errors that can occur in a task"""

    build = "BUILD_ERROR"
    execution = "EXECUTION_ERROR"
    internal = "INTERNAL_ERROR"


class OrderingFields(str, enum.Enum):
    """Model fields ordering is allowed on for list"""

    creation_date = "creation_date"
    start_date = "start_date"
    end_date = "end_date"

    @classmethod
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        else:
            return True


class Permission(schemas._PydanticConfig):
    """Permissions of a task"""

    public: bool
    authorized_ids: List[str]


class Permissions(schemas._PydanticConfig):
    """Permissions structure stored in various asset types."""

    process: Permission


class _Model(schemas._PydanticConfig, abc.ABC):
    """Asset creation specification base class."""

    # pretty print
    def __str__(self):
        return self.model_dump_json(indent=4)

    def __repr__(self):
        return self.model_dump_json(indent=4)

    @staticmethod
    def allowed_filters() -> List[str]:
        """allowed fields to filter on"""
        return []


class DataSample(_Model):
    """Data sample"""

    key: str
    owner: str
    data_manager_keys: Optional[List[str]] = None
    path: Optional[DirectoryPath] = None
    creation_date: datetime

    type_: ClassVar[str] = schemas.Type.DataSample

    @staticmethod
    def allowed_filters() -> List[str]:
        return ["key", "owner", "compute_plan_key", "function_key", "dataset_key"]


class _File(schemas._PydanticConfig):
    """File as stored in the models"""

    checksum: str
    storage_address: UriPath


class Dataset(_Model):
    """Dataset asset"""

    key: str
    name: str
    owner: str
    permissions: Permissions
    data_sample_keys: List[str] = []
    opener: _File
    description: _File
    metadata: Dict[str, str]
    creation_date: datetime
    logs_permission: Permission

    type_: ClassVar[str] = schemas.Type.Dataset

    @staticmethod
    def allowed_filters() -> List[str]:
        return ["key", "name", "owner", "permissions", "compute_plan_key", "function_key", "data_sample_key"]


class FunctionInput(_Model):
    identifier: str
    kind: schemas.AssetKind
    optional: bool
    multiple: bool


class FunctionOutput(_Model):
    identifier: str
    kind: schemas.AssetKind
    multiple: bool


class Function(_Model):
    key: str
    name: str
    owner: str
    permissions: Permissions
    metadata: Dict[str, str]
    creation_date: datetime
    inputs: List[FunctionInput]
    outputs: List[FunctionOutput]
    status: FunctionStatus

    description: _File
    archive: _File

    type_: ClassVar[str] = schemas.Type.Function

    @staticmethod
    def allowed_filters() -> List[str]:
        return ["key", "name", "owner", "permissions", "compute_plan_key", "dataset_key", "data_sample_key"]

    @pydantic.field_validator("inputs", mode="before")
    @classmethod
    def dict_input_to_list(cls, v):
        if isinstance(v, dict):
            # Transform the inputs dict to a list
            return [
                FunctionInput(
                    identifier=identifier,
                    kind=function_input["kind"],
                    optional=function_input["optional"],
                    multiple=function_input["multiple"],
                )
                for identifier, function_input in v.items()
            ]
        else:
            return v

    @pydantic.field_validator("outputs", mode="before")
    @classmethod
    def dict_output_to_list(cls, v):
        if isinstance(v, dict):
            # Transform the outputs dict to a list
            return [
                FunctionOutput(
                    identifier=identifier, kind=function_output["kind"], multiple=function_output["multiple"]
                )
                for identifier, function_output in v.items()
            ]
        else:
            return v


class InModel(schemas._PydanticConfig):
    """In model of a task"""

    checksum: str
    storage_address: UriPath


class OutModel(schemas._PydanticConfig):
    """Out model of a task"""

    key: str
    compute_task_key: str
    address: Optional[InModel] = None
    permissions: Permissions
    owner: str
    creation_date: datetime

    type_: ClassVar[str] = schemas.Type.Model

    @staticmethod
    def allowed_filters() -> List[str]:
        return ["key", "compute_task_key", "owner", "permissions"]


class InputRef(schemas._PydanticConfig):
    identifier: str
    asset_key: Optional[str] = None
    parent_task_key: Optional[str] = None
    parent_task_output_identifier: Optional[str] = None

    # either (asset_key) or (parent_task_key, parent_task_output_identifier) must be specified
    _check_asset_key_or_parent_ref = pydantic.model_validator(mode="before")(schemas.check_asset_key_or_parent_ref)


class ComputeTaskOutput(schemas._PydanticConfig):
    """Specification of a compute task input"""

    permissions: Permissions
    is_transient: bool = Field(False, alias="transient")
    model_config = ConfigDict(populate_by_name=True)


class Task(_Model):
    key: str
    function: Function
    owner: str
    compute_plan_key: str
    metadata: Dict[str, str]
    status: ComputeTaskStatus
    worker: str
    rank: Optional[int] = None
    tag: str
    creation_date: datetime
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    error_type: Optional[TaskErrorType] = None
    inputs: List[InputRef]
    outputs: Dict[str, ComputeTaskOutput]

    type_: ClassVar[Type] = schemas.Type.Task

    @staticmethod
    def allowed_filters() -> List[str]:
        return [
            "key",
            "owner",
            "worker",
            "rank",
            "status",
            "metadata",
            "compute_plan_key",
            "function_key",
        ]


Task.model_rebuild()


class ComputePlan(_Model):
    """ComputePlan"""

    key: str
    tag: str
    name: str
    owner: str
    metadata: Dict[str, str]
    task_count: int = 0
    waiting_builder_slot_count: int = 0
    building_count: int = 0
    waiting_parent_tasks_count: int = 0
    waiting_executor_slot_count: int = 0
    executing_count: int = 0
    canceled_count: int = 0
    failed_count: int = 0
    done_count: int = 0
    failed_task_key: Optional[str] = None
    status: ComputePlanStatus
    creation_date: datetime
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    estimated_end_date: Optional[datetime] = None
    duration: Optional[int] = None
    creator: Optional[str] = None

    type_: ClassVar[str] = schemas.Type.ComputePlan

    @staticmethod
    def allowed_filters() -> List[str]:
        return [
            "key",
            "name",
            "owner",
            "worker",
            "status",
            "metadata",
            "function_key",
            "dataset_key",
            "data_sample_key",
        ]


class Performances(_Model):
    """Performances of the different compute tasks of a compute plan"""

    compute_plan_key: List[str] = []
    compute_plan_tag: List[str] = []
    compute_plan_status: List[str] = []
    compute_plan_start_date: List[datetime] = []
    compute_plan_end_date: List[datetime] = []
    compute_plan_metadata: List[dict] = []
    worker: List[str] = []
    task_key: List[str] = []
    task_rank: List[int] = []
    round_idx: List[int] = []
    identifier: List[str] = []
    performance: List[float] = []


class Organization(schemas._PydanticConfig):
    """Organization"""

    id: str
    is_current: bool
    creation_date: datetime

    type_: ClassVar[str] = schemas.Type.Organization


class OrganizationInfoConfig(schemas._PydanticConfig, extra="allow"):
    model_config = ConfigDict(protected_namespaces=())
    model_export_enabled: bool


class OrganizationInfo(schemas._PydanticConfig):
    host: AnyUrl
    organization_id: str
    organization_name: str
    config: OrganizationInfoConfig
    channel: str
    version: str
    orchestrator_version: str


class _TaskAsset(schemas._PydanticConfig):
    kind: str
    identifier: str

    @staticmethod
    def allowed_filters() -> List[str]:
        return ["identifier", "kind"]


class InputAsset(_TaskAsset):
    asset: Union[Dataset, DataSample, OutModel]
    type_: ClassVar[str] = schemas.Type.InputAsset


class OutputAsset(_TaskAsset):
    asset: Union[float, OutModel]
    type_: ClassVar[str] = schemas.Type.OutputAsset

    # Deal with remote returning the actual performance object
    @pydantic.field_validator("asset", mode="before")
    @classmethod
    def convert_remote_performance(cls, value, values):
        if values.data.get("kind") == schemas.AssetKind.performance and isinstance(value, dict):
            return value.get("performance_value")

        return value


SCHEMA_TO_MODEL = {
    schemas.Type.Task: Task,
    schemas.Type.Function: Function,
    schemas.Type.ComputePlan: ComputePlan,
    schemas.Type.DataSample: DataSample,
    schemas.Type.Dataset: Dataset,
    schemas.Type.Organization: Organization,
    schemas.Type.Model: OutModel,
    schemas.Type.InputAsset: InputAsset,
    schemas.Type.OutputAsset: OutputAsset,
    schemas.Type.FunctionOutput: FunctionOutput,
    schemas.Type.FunctionInput: FunctionInput,
}



import os
import pathlib
import tarfile


class TarSafe(tarfile.TarFile):
    """
    A safe subclass of the TarFile class for interacting with tar files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.directory = os.getcwd()

    @classmethod
    def open(cls, name=None, mode="r", fileobj=None, bufsize=tarfile.RECORDSIZE, **kwargs):
        return super().open(name, mode, fileobj, bufsize, **kwargs)

    def extract(self, member, path="", set_attrs=True, *, numeric_owner=False):
        """
        Override the parent extract method and add safety checks.
        """
        self._safetar_check()
        super().extract(member, path, set_attrs=set_attrs, numeric_owner=numeric_owner)

    def extractall(self, path=".", members=None, numeric_owner=False):
        """
        Override the parent extractall method and add safety checks.
        """
        self._safetar_check()
        super().extractall(path, members, numeric_owner=numeric_owner)

    def _safetar_check(self):
        """
        Runs all necessary checks for the safety of a tarfile.
        """
        try:
            for tarinfo in self.__iter__():
                if self._is_traversal_attempt(tarinfo=tarinfo):
                    raise TarSafeError(f"Attempted directory traversal for member: {tarinfo.name}")
                if tarinfo.issym():
                    raise TarSafeError(f"Unsupported symlink for member: {tarinfo.linkname}")
                if self._is_unsafe_link(tarinfo=tarinfo):
                    raise TarSafeError(f"Attempted directory traversal via link for member: {tarinfo.linkname}")
                if self._is_device(tarinfo=tarinfo):
                    raise TarSafeError("tarfile returns true for isblk() or ischr()")
        except Exception:
            raise

    def _is_traversal_attempt(self, tarinfo):
        if not os.path.abspath(os.path.join(self.directory, tarinfo.name)).startswith(self.directory):
            return True
        return False

    def _is_unsafe_link(self, tarinfo):
        if tarinfo.islnk():
            link_file = pathlib.Path(os.path.normpath(os.path.join(self.directory, tarinfo.linkname)))
            if not os.path.abspath(os.path.join(self.directory, link_file)).startswith(self.directory):
                return True
        return False

    def _is_device(self, tarinfo):
        return tarinfo.ischr() or tarinfo.isblk()


class TarSafeError(Exception):
    pass


open = TarSafe.open



from substra.__version__ import __version__
from substra.sdk import Client
from substra.sdk import exceptions
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.schemas import BackendType

__all__ = [
    "__version__",
    "Client",
    "exceptions",
    "BackendType",
    "schemas",
    "models",
]



from substra.sdk.backends.local.compute.spawner.base import BaseSpawner
from substra.sdk.backends.local.compute.spawner.docker import Docker
from substra.sdk.backends.local.compute.spawner.subprocess import Subprocess
from substra.sdk.schemas import BackendType

__all__ = ["BaseSpawner", "Docker", "Subprocess"]

DEBUG_SPAWNER_CHOICES = {
    BackendType.LOCAL_DOCKER: Docker,
    BackendType.LOCAL_SUBPROCESS: Subprocess,
}


def get(name, *args, **kwargs):
    """Return a Docker spawner or a Subprocess spawner"""
    return DEBUG_SPAWNER_CHOICES[name](*args, **kwargs)



from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.client import Client
from substra.sdk.schemas import BackendType
from substra.sdk.utils import retry_on_exception

__all__ = [
    "Client",
    "retry_on_exception",
    "BackendType",
    "schemas",
    "models",
]



import tarfile
import zipfile

from substra.sdk import exceptions
from substra.sdk.archive import tarsafe
from substra.sdk.archive.safezip import ZipFile


def _untar(archive, to_):
    with tarsafe.open(archive, "r:*") as tf:
        tf.extractall(to_)


def _unzip(archive, to_):
    with ZipFile(archive, "r") as zf:
        zf.extractall(to_)


def uncompress(archive, to_):
    """Uncompress tar or zip archive to destination."""
    if tarfile.is_tarfile(archive):
        _untar(archive, to_)
    elif zipfile.is_zipfile(archive):
        _unzip(archive, to_)
    else:
        raise exceptions.InvalidRequest(f"Cannot uncompress '{archive}'", 400)



import argparse
import inspect
import sys
import warnings
from pathlib import Path

import pydantic

from substra.sdk import models
from substra.sdk import schemas

local_dir = Path(__file__).parent

schemas_list = [
    schemas.DataSampleSpec,
    schemas.DatasetSpec,
    schemas.UpdateDatasetSpec,
    schemas.FunctionSpec,
    schemas.FunctionInputSpec,
    schemas.FunctionOutputSpec,
    schemas.TaskSpec,
    schemas.ComputeTaskOutputSpec,
    schemas.UpdateFunctionSpec,
    schemas.ComputePlanSpec,
    schemas.UpdateComputePlanSpec,
    schemas.UpdateComputePlanTasksSpec,
    schemas.ComputePlanTaskSpec,
    schemas.Permissions,
    schemas.PrivatePermissions,
]

models_list = [
    models.DataSample,
    models.Dataset,
    models.Task,
    models.Function,
    models.ComputePlan,
    models.Performances,
    models.Organization,
    models.Permissions,
    models.InModel,
    models.OutModel,
]


def _get_field_description(fields):
    desc = [f"{name}: {field.annotation}" for name, field in fields.items()]
    return desc


def generate_help(fh, models: bool):
    if models:
        asset_list = models_list
        title = "Models"
    else:
        asset_list = schemas_list
        title = "Schemas"

    fh.write("# Summary\n\n")

    def _create_anchor(schema):
        return "#{}".format(schema.__name__)

    for asset in asset_list:
        anchor = _create_anchor(asset)
        fh.write(f"- [{asset.__name__}]({anchor})\n")

    fh.write("\n\n")
    fh.write(f"# {title}\n\n")

    for asset in asset_list:
        anchor = _create_anchor(asset)

        fh.write(f"## {asset.__name__}\n")
        # Write the docstring
        fh.write(f"{inspect.getdoc(asset)}\n")
        # List the fields and their types
        description = _get_field_description(asset.model_fields)
        fh.write("```text\n")
        fh.write("- " + "\n- ".join(description))
        fh.write("\n```")
        fh.write("\n\n")


def write_help(path, models: bool):
    with path.open("w") as fh:
        generate_help(fh, models)


if __name__ == "__main__":
    expected_pydantic_version = "2.3.0"
    if pydantic.VERSION != expected_pydantic_version:
        warnings.warn(
            f"The documentation should be generated with the version {expected_pydantic_version} of pydantic or \
            there might be mismatches with the CI: version {pydantic.VERSION} used"
        )

    doc_dir = local_dir.parent / "references"
    default_path = doc_dir / "sdk_schemas.md"

    parser = argparse.ArgumentParser()
    parser.add_argument("--output-path", type=str, default=str(default_path.resolve()), required=False)
    parser.add_argument(
        "--models",
        action="store_true",
        help="Generate the doc for the models.\
        Default: generate for the schemas",
    )

    args = parser.parse_args(sys.argv[1:])
    write_help(Path(args.output_path), models=args.models)



import typing

from substra.sdk import exceptions


def _get_inverted_node_graph(node_graph, node_to_ignore):
    """Get the graph {node_id: nodes that depend on this one}
    Also this graph does not contain the nodes to ignore
    """
    inverted = dict()
    for node, dependencies in node_graph.items():
        if node not in node_to_ignore:
            for dependency in dependencies:
                if dependency not in node_to_ignore:
                    inverted.setdefault(dependency, [])
                    inverted[dependency].append(node)
    return inverted


def _breadth_first_traversal_rank(
    ranks: typing.Dict[str, int], inverted_node_graph: typing.Dict[str, typing.List[str]]
):
    edges = set()
    queue = [node for node in ranks]
    visited = set(queue)

    if len(queue) == 0:
        raise exceptions.InvalidRequest("missing dependency among inModels IDs, circular dependency found", 400)

    while len(queue) > 0:
        current_node = queue.pop(0)
        for child in inverted_node_graph.get(current_node, []):
            new_child_rank = max(ranks[current_node] + 1, ranks.get(child, -1))

            if new_child_rank != ranks.get(child, -1):
                # either the child has never been visited
                # or its rank has been updated and we must visit again
                ranks[child] = new_child_rank
                visited.add(child)
                queue.append(child)

            # Cycle detection
            edge = (current_node, child)
            if (edge[1], edge[0]) in edges:
                raise exceptions.InvalidRequest(
                    f"missing dependency among inModels IDs, \
                        circular dependency between {edge[0]} and {edge[1]}",
                    400,
                )
            else:
                edges.add(edge)
    return ranks


def compute_ranks(
    node_graph: typing.Dict[str, typing.List[str]],
    node_to_ignore: typing.Set[str] = None,
    ranks: typing.Dict[str, int] = None,
) -> typing.Dict[str, int]:
    """Compute the ranks of the nodes in the graph.

    Args:
        node_graph (typing.Dict[str, typing.List[str]]):
            Dict {node_id: list of nodes it depends on}.
            Node graph keys must not contain any node to ignore.
        node_to_ignore (typing.Set[str], optional): List of nodes to ignore.
            Defaults to None.
        ranks (typing.Dict[str, int]): Already computed ranks. Defaults to None.

    Raises:
        exceptions.InvalidRequest: If the node graph contains a cycle

    Returns:
        typing.Dict[str, int]: Dict { node_id : rank }
    """
    ranks = ranks or dict()
    node_to_ignore = node_to_ignore or set()

    if len(node_graph) == 0:
        return dict()

    extra_nodes = set(node_graph.keys()).intersection(node_to_ignore)
    if len(extra_nodes) > 0:
        raise ValueError(f"node_graph keys should not contain any node to ignore: {extra_nodes}")

    inverted_node_graph = _get_inverted_node_graph(node_graph, node_to_ignore)

    # Assign rank 0 to nodes without deps
    for node, dependencies in node_graph.items():
        if node not in node_to_ignore:
            actual_deps = [dep for dep in dependencies if dep not in node_to_ignore]
            if len(actual_deps) == 0:
                ranks[node] = 0

    ranks = _breadth_first_traversal_rank(ranks=ranks, inverted_node_graph=inverted_node_graph)

    return ranks



import json


def format_search_filters_for_remote(filters):
    formatted_filters = {}
    # do not process if no filters
    if filters is None:
        return formatted_filters

    for key in filters:
        # handle special cases name and metadata
        if key == "name":
            formatted_filters["match"] = filters[key]
        elif key == "metadata":
            formatted_filters["metadata"] = json.dumps(filters["metadata"]).replace(" ", "")

        else:
            # all other filters are formatted as a csv string without spaces
            values = ",".join(filters[key])
            formatted_filters[key] = values.replace(" ", "")

    return formatted_filters


def format_search_ordering_for_remote(order_by, ascending):
    if not ascending:
        return "-" + order_by
    return order_by



import contextlib
import enum
import json
import pathlib
import typing
import uuid
from typing import Dict
from typing import List
from typing import Optional

import pydantic
from pydantic import ConfigDict
from pydantic.fields import Field

from substra.sdk import utils

_SERVER_NAMES = {
    "dataset": "data_manager",
    "summary_task": "task",
}


class BackendType(str, enum.Enum):
    REMOTE = "remote"
    LOCAL_DOCKER = "docker"
    LOCAL_SUBPROCESS = "subprocess"


class StaticInputIdentifier(str, enum.Enum):
    chainkeys = "chainkeys"
    datasamples = "datasamples"
    opener = "opener"

    @classmethod
    def values(cls):
        return set(item.value for item in cls)

    @classmethod
    def has_value(cls, value):
        return value in cls.values()


class AssetKind(str, enum.Enum):
    model = "ASSET_MODEL"
    performance = "ASSET_PERFORMANCE"
    data_manager = "ASSET_DATA_MANAGER"
    data_sample = "ASSET_DATA_SAMPLE"


class Type(enum.Enum):
    Function = "function"
    FunctionOutput = "function_output"
    FunctionInput = "function_input"
    DataSample = "data_sample"
    Dataset = "dataset"
    Model = "model"
    ComputePlan = "compute_plan"
    Organization = "organization"
    Task = "task"
    InputAsset = "input_asset"
    OutputAsset = "output_asset"

    def to_server(self):
        """Returns the name used to identify the asset on the backend."""
        name = self.value
        return _SERVER_NAMES.get(name, name)

    def to_asset_kind(self):
        """Returns the name used to identify the asset on the backend."""
        if self.value == self.Dataset.value:
            return AssetKind.data_manager
        else:
            return AssetKind[self.value]

    def __str__(self):
        return self.name


class _PydanticConfig(pydantic.BaseModel):
    """Shared configuration for all schemas here"""

    model_config = ConfigDict(extra="ignore")


class _Spec(_PydanticConfig):
    """Asset creation specification base class."""

    # pretty print
    def __str__(self):
        return self.model_dump_json(indent=4)

    def __repr__(self):
        return self.model_dump_json(indent=4)

    class Meta:
        file_attributes = None

    def is_many(self):
        return False

    @contextlib.contextmanager
    def build_request_kwargs(self):
        # TODO should be located in the backends/remote module
        # Serialize and deserialize to prevent errors eg with pathlib.Path
        data = json.loads(self.model_dump_json(exclude_unset=True))
        if self.Meta.file_attributes:
            with utils.extract_files(data, self.Meta.file_attributes) as (data, files):
                yield (data, files)
        else:
            yield data, None

    @staticmethod
    def compute_key() -> str:
        return str(uuid.uuid4())


class Permissions(_PydanticConfig):
    """Specification for permissions. If public is False,
    give the list of authorized ids.
    """

    public: bool
    authorized_ids: typing.List[str]  # List of authorized organization ids if private


class PrivatePermissions(_PydanticConfig):
    """Specification for private permissions. Only the organizations whose
    ids are in authorized_ids can access the asset.
    """

    authorized_ids: typing.List[str]  # List of authorized organization ids


class DataSampleSpec(_Spec):
    """Specification to create one or many data samples
    To create one data sample, use the 'path' field, otherwise use
    the 'paths' field.
    """

    path: Optional[pathlib.Path] = None  # Path to the data sample if only one
    paths: Optional[List[pathlib.Path]] = None  # Path to the data samples if several
    data_manager_keys: typing.List[str]

    type_: typing.ClassVar[Type] = Type.DataSample

    def is_many(self):
        return self.paths and len(self.paths) > 0

    @pydantic.field_validator("paths")
    @classmethod
    def resolve_paths(cls, v: List[pathlib.Path]) -> List[pathlib.Path]:
        """Resolve given paths."""
        if v is None:
            raise ValueError("'paths' cannot be set to None.")

        return [p.resolve() for p in v]

    @pydantic.field_validator("path")
    @classmethod
    def resolve_path(cls, v: pathlib.Path) -> pathlib.Path:
        """Resolve given path."""
        if v is None:
            raise ValueError("'path' cannot be set to None.")

        return v.resolve()

    @pydantic.model_validator(mode="before")
    @classmethod
    def exclusive_paths(cls, values: typing.Any) -> typing.Any:
        """Check that one and only one path(s) field is defined."""
        if "paths" in values and "path" in values:
            raise ValueError("'path' and 'paths' fields are exclusive.")
        if "paths" not in values and "path" not in values:
            raise ValueError("'path' or 'paths' field must be set.")
        return values

    @contextlib.contextmanager
    def build_request_kwargs(self, local):
        # redefine kwargs builder to handle the local paths
        # Serialize and deserialize to prevent errors eg with pathlib.Path
        data = json.loads(self.model_dump_json(exclude_unset=True))
        if local:
            with utils.extract_data_sample_files(data) as (data, files):
                yield (data, files)
        else:
            yield data, None


def check_asset_key_or_parent_ref(cls, values):
    """Check that either (asset key) or (parent_task_key, parent_task_output_identifier) are set, but not both."""

    has_asset_key = bool(dict(values).get("asset_key"))
    has_parent = bool(dict(values).get("parent_task_key")) and bool(dict(values).get("parent_task_output_identifier"))

    if has_asset_key != has_parent:  # xor
        return values

    raise ValueError("either asset_key or both (parent_task_key, parent_task_output_identifier) must be provided")


class InputRef(_PydanticConfig):
    """Specification of a compute task input"""

    identifier: str
    asset_key: Optional[str] = None
    parent_task_key: Optional[str] = None
    parent_task_output_identifier: Optional[str] = None

    # either (asset_key) or (parent_task_key, parent_task_output_identifier) must be specified
    _check_asset_key_or_parent_ref = pydantic.model_validator(mode="before")(check_asset_key_or_parent_ref)


class ComputeTaskOutputSpec(_PydanticConfig):
    """Specification of a compute task output"""

    permissions: Permissions
    is_transient: Optional[bool] = Field(False, alias="transient")
    model_config = ConfigDict(populate_by_name=True)


class ComputePlanTaskSpec(_Spec):
    """Specification of a compute task inside a compute plan specification

    note : metadata field does not accept strings containing '__' as dict key
    """

    task_id: str
    function_key: str
    worker: str
    tag: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    inputs: Optional[List[InputRef]] = None
    outputs: Optional[Dict[str, ComputeTaskOutputSpec]] = None


class _BaseComputePlanSpec(_Spec):
    key: str
    tasks: Optional[List[ComputePlanTaskSpec]] = None


class ComputePlanSpec(_BaseComputePlanSpec):
    """Specification for creating a compute plan

    note : metadata field does not accept strings containing '__' as dict key
    """

    key: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    tag: Optional[str] = None
    name: str
    metadata: Optional[Dict[str, str]] = None

    type_: typing.ClassVar[Type] = Type.ComputePlan

    @contextlib.contextmanager
    def build_request_kwargs(self):
        # default values are not dumped when `exclude_unset` flag is enabled,
        # this is why we need to reimplement this custom function.
        data = json.loads(self.model_dump_json(exclude_unset=True))
        data["key"] = self.key
        yield data, None


class UpdateComputePlanTasksSpec(_BaseComputePlanSpec):
    """Specification for updating a compute plan's tasks"""

    pass


class UpdateComputePlanSpec(_Spec):
    """Specification for updating a compute plan"""

    name: str

    type_: typing.ClassVar[Type] = Type.ComputePlan


class DatasetSpec(_Spec):
    """Specification for creating a dataset

    note : metadata field does not accept strings containing '__' as dict key
    """

    name: str
    data_opener: pathlib.Path  # Path to the data opener
    description: pathlib.Path  # Path to the description file
    permissions: Permissions
    metadata: Optional[Dict[str, str]] = None
    logs_permission: Permissions

    type_: typing.ClassVar[Type] = Type.Dataset

    class Meta:
        file_attributes = (
            "data_opener",
            "description",
        )


class UpdateDatasetSpec(_Spec):
    """Specification for updating a dataset"""

    name: str

    type_: typing.ClassVar[Type] = Type.Dataset


class FunctionInputSpec(_Spec):
    identifier: str
    multiple: bool
    optional: bool
    kind: AssetKind

    @pydantic.model_validator(mode="before")
    @classmethod
    def _check_identifiers(cls, values):
        """Checks that the multiplicity and the optionality of a data manager is always set to False"""
        if values["kind"] == AssetKind.data_manager:
            if values["multiple"]:
                raise ValueError("Data manager input can't be multiple.")
            if values["optional"]:
                raise ValueError("Data manager input can't be optional.")
            if values["identifier"] != StaticInputIdentifier.opener:
                raise ValueError(
                    f"Data manager input identifier must be `{StaticInputIdentifier.opener}` "
                    f"but was set to {values['identifier']}"
                )
        elif values["kind"] == AssetKind.data_sample:
            if values["identifier"] != StaticInputIdentifier.datasamples:
                raise ValueError(
                    f"Data sample input identifier must be `{StaticInputIdentifier.datasamples}` "
                    f"but was set to {values['identifier']}"
                )
        else:
            if StaticInputIdentifier.has_value(values["identifier"]):
                raise ValueError(
                    f"Inputs of kind {values['kind']} cannot have an identifier among the flowing: "
                    f"{StaticInputIdentifier.values()}"
                )

        return values


class FunctionOutputSpec(_Spec):
    identifier: str
    kind: AssetKind
    multiple: bool

    @pydantic.model_validator(mode="before")
    @classmethod
    def _check_performance(cls, values):
        """Checks that the performance is always set to False"""
        if values == AssetKind.performance and values["multiple"]:
            raise ValueError("Performance can't be multiple.")

        return values


class FunctionSpec(_Spec):
    """Specification for creating an function

    note : metadata field does not accept strings containing '__' as dict key
    """

    name: str
    description: pathlib.Path
    file: pathlib.Path
    permissions: Permissions
    metadata: Optional[Dict[str, str]] = None
    inputs: Optional[List[FunctionInputSpec]] = None
    outputs: Optional[List[FunctionOutputSpec]] = None

    type_: typing.ClassVar[Type] = Type.Function

    @pydantic.field_validator("inputs")
    @classmethod
    def _check_inputs(cls, v):
        inputs = v or []
        identifiers = {value.identifier for value in inputs}
        if len(identifiers) != len(inputs):
            raise ValueError("Several function inputs cannot have the same identifier.")
        return v

    @pydantic.field_validator("outputs")
    @classmethod
    def _check_outputs(cls, v):
        outputs = v or []
        identifiers = {value.identifier for value in outputs}
        if len(identifiers) != len(outputs):
            raise ValueError("Several function outputs cannot have the same identifier.")
        return v

    @contextlib.contextmanager
    def build_request_kwargs(self):
        # TODO should be located in the backends/remote module
        # Serialize and deserialize to prevent errors eg with pathlib.Path
        data = json.loads(self.model_dump_json(exclude_unset=True))

        # Computed fields using `@property` are not dumped when `exclude_unset` flag is enabled,
        # this is why we need to reimplement this custom function.
        data["inputs"] = (
            {input.identifier: input.model_dump(exclude={"identifier"}) for input in self.inputs}
            if self.inputs
            else dict()
        )
        data["outputs"] = (
            {output.identifier: output.model_dump(exclude={"identifier"}) for output in self.outputs}
            if self.outputs
            else dict()
        )

        if self.Meta.file_attributes:
            with utils.extract_files(data, self.Meta.file_attributes) as (data, files):
                yield (data, files)
        else:
            yield data, None

    class Meta:
        file_attributes = (
            "file",
            "description",
        )


class UpdateFunctionSpec(_Spec):
    """Specification for updating an function"""

    name: str

    type_: typing.ClassVar[Type] = Type.Function


class TaskSpec(_Spec):
    key: str = pydantic.Field(default_factory=lambda: str(uuid.uuid4()))
    tag: Optional[str] = None
    compute_plan_key: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    function_key: str
    worker: str
    rank: Optional[int] = None
    inputs: Optional[List[InputRef]] = None
    outputs: Optional[Dict[str, ComputeTaskOutputSpec]] = None

    type_: typing.ClassVar[Type] = Type.Task

    @contextlib.contextmanager
    def build_request_kwargs(self):
        # default values are not dumped when `exclude_unset` flag is enabled,
        # this is why we need to reimplement this custom function.
        data = json.loads(self.model_dump_json(exclude_unset=True))
        data["key"] = self.key
        data["inputs"] = [input.model_dump() for input in self.inputs] if self.inputs else []
        data["outputs"] = {k: v.model_dump(by_alias=True) for k, v in self.outputs.items()} if self.outputs else {}
        yield data, None

    @classmethod
    def from_compute_plan(cls, compute_plan_key: str, rank: int, spec: ComputePlanTaskSpec) -> "TaskSpec":
        return TaskSpec(
            key=spec.task_id,
            function_key=spec.function_key,
            inputs=spec.inputs,
            outputs=spec.outputs,
            tag=spec.tag,
            compute_plan_key=compute_plan_key,
            rank=rank,
            metadata=spec.metadata,
            worker=spec.worker,
        )



import logging
import os
import pathlib
import re
import shutil
import string
import subprocess
import sys
import tempfile
import typing

from substra.sdk.archive import uncompress
from substra.sdk.backends.local.compute.spawner.base import VOLUME_INPUTS
from substra.sdk.backends.local.compute.spawner.base import BaseSpawner
from substra.sdk.backends.local.compute.spawner.base import ExecutionError
from substra.sdk.backends.local.compute.spawner.base import write_command_args_file

logger = logging.getLogger(__name__)

PYTHON_SCRIPT_REGEX = r"(?<=\")([^\"]*\.py)(?=\")"
METHOD_REGEX = r"\"\-\-function-name\"\,\s*\"([^\"]*)\""


def _get_entrypoint_from_dockerfile(tmpdir):
    """
    Extracts the .py script and the function name to execute in
    an ENTRYPOINT line of a Dockerfile, located in tmpdir.
    For instance if the line `ENTRYPOINT ["python3", "function.py", "--function-name", "train"]` is in the Dockerfile,
     `function.py`, `train` is extracted.
    """
    valid_example = (
        """The entry point should be specified as follow: """
        """``ENTRYPOINT ["<executor>", "<function_file.py>", "--function-name", "<method name>"]"""
    )
    with open(tmpdir / "Dockerfile") as f:
        for line in f:
            if "ENTRYPOINT" not in line:
                continue

            script_name = re.findall(PYTHON_SCRIPT_REGEX, line)
            if len(script_name) != 1:
                raise ExecutionError("Couldn't extract script from ENTRYPOINT line in Dockerfile", valid_example)

            function_name = re.findall(METHOD_REGEX, line)
            if len(function_name) != 1:
                raise ExecutionError("Couldn't extract method name from ENTRYPOINT line in Dockerfile", valid_example)

            return script_name[0], function_name[0]

    raise ExecutionError("Couldn't get entrypoint in Dockerfile", valid_example)


def _get_command_args(
    function_name: str, args_template: typing.List[string.Template], local_volumes: typing.Dict[str, str]
) -> typing.List[str]:
    args = ["--function-name", str(function_name)]
    args += [tpl.substitute(**local_volumes) for tpl in args_template]
    return args


def _symlink_data_samples(data_sample_paths: typing.Dict[str, pathlib.Path], dest_dir: str):
    """Create a symbolic link to "move" the data samples
    to the data directory.

    Args:
        data_sample_paths (typing.Dict[str, pathlib.Path]): Paths to the samples
        dest_dir (str): Temp data directory
    """
    # Check if there are already data samples in the dest dir (testtasks are executed in 2 parts)
    sample_key = next(iter(data_sample_paths))
    if (pathlib.Path(dest_dir) / sample_key).exists():
        return
    # copy the whole tree but using hard link to be fast and not use too much place
    for sample_key, sample_path in data_sample_paths.items():
        dest_path = pathlib.Path(dest_dir) / sample_key
        shutil.copytree(sample_path, dest_path, copy_function=os.symlink)


class Subprocess(BaseSpawner):
    """Wrapper to execute a command in a python process."""

    def __init__(self, local_worker_dir: pathlib.Path):
        super().__init__(local_worker_dir=local_worker_dir)

    def spawn(
        self,
        name,
        archive_path,
        command_args_tpl: typing.List[string.Template],
        data_sample_paths: typing.Optional[typing.Dict[str, pathlib.Path]],
        local_volumes,
        envs,
    ):
        """Spawn a python process (blocking)."""
        with tempfile.TemporaryDirectory(dir=self._local_worker_dir) as function_dir:
            with tempfile.TemporaryDirectory(dir=function_dir) as args_dir:
                function_dir = pathlib.Path(function_dir)
                args_dir = pathlib.Path(args_dir)
                uncompress(archive_path, function_dir)
                script_name, function_name = _get_entrypoint_from_dockerfile(function_dir)

                args_file = args_dir / "arguments.txt"

                py_command = [sys.executable, str(function_dir / script_name), f"@{args_file}"]
                py_command_args = _get_command_args(function_name, command_args_tpl, local_volumes)
                write_command_args_file(args_file, py_command_args)

                if data_sample_paths is not None and len(data_sample_paths) > 0:
                    _symlink_data_samples(data_sample_paths, local_volumes[VOLUME_INPUTS])

                # Catching error and raising to be ISO to the docker local backend
                # Don't capture the output to be able to use pdb
                try:
                    subprocess.run(py_command, capture_output=False, check=True, cwd=function_dir, env=envs)
                except subprocess.CalledProcessError as e:
                    raise ExecutionError(e)



import contextlib
import copy
import functools
import io
import logging
import ntpath
import os
import re
import time
import uuid
import zipfile

from substra.sdk import exceptions
from substra.sdk import models


def path_leaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


@contextlib.contextmanager
def extract_files(data, file_attributes):
    data = copy.deepcopy(data)

    paths = {}
    for attr in file_attributes:
        try:
            paths[attr] = data[attr]
        except KeyError:
            raise exceptions.LoadDataException(f"The '{attr}' attribute is missing.")
        del data[attr]

    files = {}
    for k, f in paths.items():
        if not os.path.exists(f):
            raise exceptions.LoadDataException(f"The '{k}' attribute file ({f}) does not exist.")
        files[k] = open(f, "rb")

    try:
        yield (data, files)
    finally:
        for f in files.values():
            f.close()


def zip_folder(fp, path):
    zipf = zipfile.ZipFile(fp, "w", zipfile.ZIP_DEFLATED)
    for root, _, files in os.walk(path):
        for f in files:
            abspath = os.path.join(root, f)
            archive_path = os.path.relpath(abspath, start=path)
            zipf.write(abspath, arcname=archive_path)
    zipf.close()


def zip_folder_in_memory(path):
    fp = io.BytesIO()
    zip_folder(fp, path)
    fp.seek(0)
    return fp


@contextlib.contextmanager
def extract_data_sample_files(data):
    # handle data sample specific case; paths and path cases
    data = copy.deepcopy(data)

    folders = {}
    if data.get("path"):
        attr = "path"
        folders[attr] = data[attr]
        del data[attr]

    if data.get("paths"):  # field is set and is not None/empty
        for p in data["paths"]:
            folders[path_leaf(p)] = p
        del data["paths"]

    files = {}
    for k, f in folders.items():
        if not os.path.isdir(f):
            raise exceptions.LoadDataException(f"Paths '{f}' is not an existing directory")
        files[k] = zip_folder_in_memory(f)

    try:
        yield (data, files)
    finally:
        for f in files.values():
            f.close()


def _check_metadata_search_filter(filter):
    if not isinstance(filter, dict):
        raise exceptions.FilterFormatError(
            "Cannot load filters. Please review the documentation, metadata filter should be a list of dict."
            "But one passed elements is not."
        )

    if "key" not in filter.keys() or "type" not in filter.keys():
        raise exceptions.FilterFormatError("Each metadata filter, must contains both `key` and `type` as key.")

    if filter["type"] not in ("is", "contains", "exists"):
        raise exceptions.FilterFormatError(
            "Each metadata filter `type` filed value must be `is`, `contains` or `exists`"
        )

    if filter["type"] in ("is", "contains"):
        if "value" not in filter.keys():
            raise exceptions.FilterFormatError(
                "For each metadata filter, if `type` value is `is` or `contains`, the filter should also contain the "
                "`value` key."
            )

        if not isinstance(filter.get("value"), str):
            raise exceptions.FilterFormatError(
                "For each metadata filter, if a `value` is passed, it should be a string."
            )


def _check_metadata_search_filters(filters):
    if not isinstance(filters, list):
        raise exceptions.FilterFormatError(
            "Cannot load filters. Please review the documentation, metadata filter should be a list of dict."
        )

    for filter in filters:
        _check_metadata_search_filter(filter)


def check_and_format_search_filters(asset_type, filters):  # noqa: C901
    # do not check if no filters
    if filters is None:
        return filters

    # check filters structure
    if not isinstance(filters, dict):
        raise exceptions.FilterFormatError(
            "Cannot load filters. Please review the documentation, filters should be a dict"
        )

    # retrieving asset allowed fields to filter on
    allowed_filters = models.SCHEMA_TO_MODEL[asset_type].allowed_filters()

    # for each attribute (key) to filter on
    for key in filters:
        # check that key is a valid filter
        if key not in allowed_filters:
            raise exceptions.NotAllowedFilterError(
                f"Cannot filter on {key}. Please review the documentation, filtering allowed only on {allowed_filters}"
            )
        elif key == "name":
            if not isinstance(filters[key], str):
                raise exceptions.FilterFormatError(
                    """Cannot load filters. Please review the documentation, 'name' filter is partial match in remote,
                    exact match in local, value should be str"""
                )
        elif key == "metadata":
            _check_metadata_search_filters(filters[key])

        # all other filters should be a list, throw an error if not
        elif not isinstance(filters[key], list):
            raise exceptions.FilterFormatError(
                "Cannot load filters. Please review the documentation, filters values should be a list"
            )
        # handle default case (List)
        else:
            # convert all keys to str, needed for rank as user can give int, can prevent errors if user doesn't give str
            filters[key] = [str(v) for v in filters[key]]

    return filters


def check_search_ordering(order_by):
    if order_by is None:
        return
    elif not models.OrderingFields.__contains__(order_by):
        raise exceptions.OrderingFormatError(
            f"Please review the documentation, ordering is available only on {list(models.OrderingFields.__members__)}"
        )


def retry_on_exception(exceptions, timeout=300):
    """Retry function in case of exception(s).

    Args:
        exceptions (list): list of exception types that trigger a retry
        timeout (int, optional): timeout in seconds

    Example:
        ```python
        from substra.sdk import exceptions, retry_on_exception

        def my_function(arg1, arg2):
            pass

        retry = retry_on_exception(
                    exceptions=(exceptions.RequestTimeout),
                    timeout=300,
                )
        retry(my_function)(arg1, arg2)
        ```
    """

    def _retry(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            delay = 1
            backoff = 2
            tstart = time.time()

            while True:
                try:
                    return f(*args, **kwargs)

                except exceptions:
                    if timeout is not False and time.time() - tstart > timeout:
                        raise
                    logging.warning(f"Function {f.__name__} failed: retrying in {delay}s")
                    time.sleep(delay)
                    delay *= backoff

        return wrapper

    return _retry


def response_get_destination_filename(response):
    """Get filename from content-disposition header."""
    disposition = response.headers.get("content-disposition")
    if not disposition:
        return None
    filenames = re.findall("filename=(.+)", disposition)
    if not filenames:
        return None
    filename = filenames[0]
    filename = filename.strip("'\"")
    return filename


def is_valid_uuid(value):
    try:
        uuid.UUID(value)
        return True
    except ValueError:
        return False



import hashlib


class Hasher:
    def __init__(self, values=None):
        self._h = hashlib.sha256()
        if values:
            for v in values:
                self.update(v)

    def update(self, v):
        if isinstance(v, str):
            v = v.encode("utf-8")
        self._h.update(v)

    def compute(self):
        return self._h.hexdigest()



import abc
import pathlib
import string
import typing

VOLUME_CLI_ARGS = "_VOLUME_CLI_ARGS"
VOLUME_INPUTS = "_VOLUME_INPUTS"
VOLUME_OUTPUTS = "_VOLUME_OUTPUTS"


class BuildError(Exception):
    """An error occurred during the build of the function"""

    pass


class ExecutionError(Exception):
    """An error occurred during the execution of the compute task"""

    pass


class BaseSpawner(abc.ABC):
    """Base wrapper to execute a command"""

    def __init__(self, local_worker_dir: pathlib.Path):
        self._local_worker_dir = local_worker_dir

    @abc.abstractmethod
    def spawn(
        self,
        name,
        archive_path,
        command_args_tpl: typing.List[string.Template],
        data_sample_paths: typing.Optional[typing.Dict[str, pathlib.Path]],
        local_volumes,
        envs,
    ):
        """Execute archive in a contained environment."""
        raise NotImplementedError


def write_command_args_file(args_file: pathlib.Path, command_args: typing.List[str]) -> None:
    """Write the substra-tools command line arguments to a file.

    The format uses one line per argument. See
    https://docs.python.org/3/library/argparse.html#fromfile-prefix-chars
    """
    with open(args_file, "w") as f:
        for item in command_args:
            f.write(item + "\n")
